from __future__ import annotations

import ctypes
import subprocess

from ctypes import c_char_p, c_int
from pathlib import PosixPath
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import os


# Check if a command exists
def command_exists(cmd: str) -> bool:
    return subprocess.call(["which", cmd], stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0  # noqa: S603, S607


# Get default actions using xdg-mime and gio
def get_default_actions(path: os.PathLike | str) -> list[str]:
    try:
        mime_type = subprocess.check_output(["xdg-mime", "query", "filetype", path], text=True).strip()  # noqa: S603, S607
        actions = subprocess.check_output(["gio", "mime", mime_type], text=True).splitlines()  # noqa: S603, S607
    except Exception:  # noqa: BLE001
        return []
    else:
        return [action.split(":")[1].strip() for action in actions if action.startswith("default:")]


# ctypes method 1: Using GTK
def context_menu_gtk(path: os.PathLike | str) -> bool:
    try:
        libgtk = ctypes.CDLL("libgtk-3.so.0")
        libgtk.gtk_init(None, None)
        menu = libgtk.gtk_menu_new()
        actions = get_default_actions(path)
        if not actions:
            return False
        items = [libgtk.gtk_menu_item_new_with_label(action.encode()) for action in actions]
        for item in items:
            libgtk.gtk_menu_shell_append(menu, item)
        libgtk.gtk_widget_show_all(menu)
        status_icon = libgtk.gtk_status_icon_new()
        libgtk.gtk_status_icon_set_visible(status_icon, 1)
        libgtk.gtk_status_icon_set_from_icon_name(status_icon, b"system-file-manager")
        libgtk.gtk_status_icon_set_tooltip_text(status_icon, b"Context Menu")
        libgtk.gtk_main()
    except Exception:  # noqa: BLE001
        return False
    else:
        return True


# ctypes method 2: Using Qt
def context_menu_qt(path: os.PathLike | str) -> bool:
    try:
        libQt5 = ctypes.CDLL("libQt5Widgets.so.5")
        argc, argv = c_int(1), (c_char_p * 1)(b"ctypes_menu")
        app, menu = libQt5.QApplication_New(argc, argv), libQt5.QMenu_New()
        actions = get_default_actions(path)
        if not actions:
            return False
        for action in actions:
            libQt5.QMenu_addAction(menu, action.encode())
        libQt5.QMenu_exec(menu)
        libQt5.qApp.exec()  # type: ignore[attr-defined]
    except Exception:  # noqa: BLE001
        return False
    else:
        return True


# ctypes method 3: Using wxWidgets
def context_menu_wxwidgets(path: os.PathLike | str) -> bool:
    try:
        libwx = ctypes.CDLL("libwx_baseu-3.0.so.0")
        argc, argv = c_int(1), (c_char_p * 1)(b"ctypes_menu")
        libwx.wxEntryStart(argc, argv)
        app = libwx.wxTheApp()
        frame = libwx.wxFrame_New(None, -1, b"Context Menu", (50, 50), (450, 340))
        menu, menu_bar = libwx.wxMenu_New(), libwx.wxMenuBar_New()
        actions = get_default_actions(path)
        if not actions:
            return False
        for i, action in enumerate(actions):
            libwx.wxMenu_Append(menu, i + 1, action.encode())
        libwx.wxMenuBar_Append(menu_bar, menu, b"File")
        libwx.wxFrame_SetMenuBar(frame, menu_bar)
        libwx.wxEntry()
    except Exception:  # noqa: BLE001
        return False
    else:
        return True


def context_menu_zenity(path: str):
    if not command_exists("zenity"):
        return False
    try:
        actions = get_default_actions(path)
        if not actions:
            return False
        result = subprocess.run(
            ["zenity", "--list", "--title=Context Menu", "--column=Options", *actions],  # noqa: S603, S607
            stdout=subprocess.PIPE,
            text=True,
            check=False,
        )
        if result.returncode == 0:
            choice = result.stdout.strip()
            if choice:
                subprocess.run(["zenity", "--info", f"--text=You chose '{choice}' for '{path}'"], check=False)  # noqa: S603, S607
    except Exception as e:  # noqa: BLE001
        print(f"Zenity method failed: {e}")
        return False
    else:
        return result.returncode == 0


def context_menu_yad(path: str):
    if not command_exists("yad"):
        return False
    try:
        actions = get_default_actions(path)
        if not actions:
            return False
        result = subprocess.run(
            ["yad", "--list", "--title=Context Menu", "--column=Options", *actions],  # noqa: S607, S603
            stdout=subprocess.PIPE,
            text=True,
            check=False,
        )
        if result.returncode == 0:
            choice = result.stdout.strip()
            if choice:
                subprocess.run(["yad", "--info", f"--text=You chose '{choice}' for '{path}'"], check=False)  # noqa: S607, S603
    except Exception as e:  # noqa: BLE001
        print(f"Yad method failed: {e}")
        return False
    else:
        return result.returncode == 0


def context_menu_dmenu(path: str):  # noqa: ANN201
    if not command_exists("dmenu") or not command_exists("notify-send"):
        return False
    try:
        actions = get_default_actions(path)
        if not actions:
            return False
        process = subprocess.Popen(
            ["dmenu", "-p", "Choose an option:"],  # noqa: S607, S603
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            text=True,
        )
        result, _ = process.communicate(input="\n".join(actions))
        if process.returncode == 0:
            choice = result.strip()
            if choice:
                subprocess.run(["notify-send", f"You chose '{choice}' for '{path}'"], check=False)  # noqa: S603, S607
    except Exception as e:  # noqa: BLE001
        print(f"Spawning context menu through dmenu failed: {e.__class__.__name__}: {e}")
        return False
    else:
        return process.returncode == 0


def show_context_menu(path: str):
    abs_path = PosixPath(path).resolve()

    def cmd_exists(cmd) -> bool:
        return subprocess.call(["which", cmd], stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0  # noqa: S603, S607

    def get_actions(p) -> list[str]:
        return [
            a.split(":")[1].strip()
            for a in subprocess.check_output(
                ["gio", "mime", subprocess.check_output(["xdg-mime", "query", "filetype", p], text=True).strip()],  # noqa: S607, S603
                text=True,
            ).splitlines()
            if a.startswith("default:")
        ]

    methods = [
        lambda p: cmd_exists("nautilus") and not subprocess.run(["nautilus", "--select", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("dolphin") and not subprocess.run(["dolphin", "--select", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("thunar") and not subprocess.run(["thunar", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("nemo") and not subprocess.run(["nemo", "--no-desktop", "--browser", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("caja") and not subprocess.run(["caja", "--browser", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("pcmanfm") and not subprocess.run(["pcmanfm", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("konqueror") and not subprocess.run(["konqueror", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("spacefm") and not subprocess.run(["spacefm", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("rox") and not subprocess.run(["rox", p], check=False).returncode,  # noqa: S603, S607
        lambda p: cmd_exists("krusader") and not subprocess.run(["krusader", p], check=False).returncode,  # noqa: S603, S607
        context_menu_zenity,
        context_menu_yad,
        context_menu_dmenu,
        context_menu_gtk,
        context_menu_qt,
        context_menu_wxwidgets,
    ]
    abs_path = PosixPath(path).resolve()
    for method in methods:
        if method(str(abs_path)):
            return
    print("All methods failed to display the context menu.")


if __name__ == "__main__":
    show_context_menu(r"C:\Program Files (x86)\Steam\steamapps\common\swkotor\modules\danm15.rim")



"""KEY/BIF writer: abstract writer and helpers for building key/bif archives."""

from __future__ import annotations

import os
import struct
import time

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import BinaryIO

from pykotor.resource.type import ResourceType


@dataclass
class Entry:
    file_name: str
    files: list[str]
    file_size: int


class KEYDataWriter(ABC):
    @abstractmethod
    def size(self) -> int:
        """Get the size of this data file, to write it into the KEY file.

        Returns:
            The total size of this data file
        """

    @abstractmethod
    def add(
        self,
        data: BinaryIO,
        restype: ResourceType,
    ) -> None:
        """Add a stream of the specified type to the data writer.

        Args:
            data: the data to write
            type: the type of this data
        """


class KEYWriter:
    """Writes KEY (Keyfile) files.

    KEY files index BIF/BZF archives and provide resource lookup tables. This writer
    creates KEY files by collecting BIF entries and their contained resources.

    References:
    ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorKEY/KEY.cs (KEY structure)
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/KEYObject.ts (KEY loading)


        Note: KEY writing is uncommon in vendor implementations; most tools only read KEY files.
        PyKotor's KEYWriter is primarily for modding and tooling purposes.

    Missing Features:
    ----------------
        - ResRef lowercasing (vendor implementations lowercase ResRefs)
        - Resource ID decomposition (vendor implementations decompose resource IDs)
        - BZF compression support (vendor implementations handle compressed BIFs)
    """

    def __init__(self):
        self._entries: list[Entry] = []

    def add_bif(
        self,
        file_name: str,
        files: list[str],
        size: int,
    ) -> None:
        """Add a reference to a specific BIF/BZF file to this KEY file.

        Args:
            file_name: the filename of the specific BIF/BZF file
            files: the files contained in this BIF/BZF file
            size: total size of the bif file
        """
        self._entries.append(Entry(file_name, files, size))

    def write(
        self,
        write_stream: BinaryIO,
    ) -> None:
        """Write the collected information as KEY file to the specified WriteStream.

        Args:
            write_stream: The stream to write to

        References:
        ----------

    Derivations and Other Implementations:
    -------------------------------------
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorKEY/KEYReader.cs (KEY reading)
        KEY file format: 8-byte signature, BIF count, resource count, offsets, timestamps

        """
        # Write header

        write_stream.write(struct.pack(">4s4s", b"KEY ", b"V1  "))

        # Number of BIF/BZF files
        write_stream.write(struct.pack("<I", len(self._entries)))

        total_count: int = sum(len(entry.files) for entry in self._entries)
        total_filename_size: int = sum(len(entry.file_name) for entry in self._entries)

        # Number of resources in all BIF/BZF files
        write_stream.write(struct.pack("<I", total_count))

        # Constant offset to the file table
        write_stream.write(struct.pack("<I", 64))

        # Calculate the key table offset
        key_table_offset: int = 64 + len(self._entries) * 12 + total_filename_size

        # Offset to the keytable
        write_stream.write(struct.pack("<I", key_table_offset))

        # Write the creation time of the file
        now: time.struct_time = time.localtime()
        write_stream.write(struct.pack("<II", now.tm_year, now.tm_yday))

        # Reserved padding
        write_stream.write(b"\0" * 32)

        # Write file table
        filename_offset: int = 64 + len(self._entries) * 12
        for entry in self._entries:
            write_stream.write(struct.pack("<III", entry.file_size, filename_offset, len(entry.file_name)))
            filename_offset += len(entry.file_name)

        # Write file name table
        for entry in self._entries:
            write_stream.write(entry.file_name.encode("ascii"))

        # Write Key table
        for x, entry in enumerate(self._entries):
            for y, file in enumerate(entry.files):
                write_stream.write(os.path.basename(file).encode("ascii"))  # noqa: PTH119
                write_stream.write(struct.pack("<HI", ResourceType.from_extension(file.split(".")[-1]).type_id, (x << 20) + y))



"""Module entry point for PyKotor CLI.

This allows PyKotor to be run as:
    python -m pykotor
"""

from __future__ import annotations

from pykotor.cli.__main__ import main

if __name__ == "__main__":
    import sys

    sys.exit(main())



"""DXT compression: RGB/RGBA to DXT1/DXT3/DXT5 block encode."""

from __future__ import annotations


def rgb_to_dxt1(
    rgb_data: bytes | bytearray,
    width: int,
    height: int,
) -> bytearray:
    dxt1_data = bytearray()
    for y in range(0, height, 4):
        for x in range(0, width, 4):
            block: list[int] = _extract_block(rgb_data, x, y, width, height, 3)
            dest = bytearray(8)
            _compress_dxt1_block(dest, block)
            dxt1_data.extend(dest)
    return dxt1_data


def rgba_to_dxt3(
    rgba_data: bytes | bytearray,
    width: int,
    height: int,
) -> bytearray:
    dxt3_data = bytearray()
    for y in range(0, height, 4):
        for x in range(0, width, 4):
            block: list[int] = _extract_block(rgba_data, x, y, width, height, 4)
            dest = bytearray(16)
            _compress_dxt3_block(dest, block)
            dxt3_data.extend(dest)
    return dxt3_data


def rgba_to_dxt5(
    rgba_data: bytes | bytearray,
    width: int,
    height: int,
) -> bytearray:
    dxt5_data = bytearray()
    for y in range(0, height, 4):
        for x in range(0, width, 4):
            block: list[int] = _extract_block(rgba_data, x, y, width, height, 4)
            dest = bytearray(16)
            _compress_dxt5_block(dest, block)
            dxt5_data.extend(dest)
    return dxt5_data


def _extract_block(
    src: bytes | bytearray,
    x: int,
    y: int,
    w: int,
    h: int,
    channels: int,
) -> list[int]:  # noqa: PLR0913
    block: list[int] = []
    for by in range(4):
        for bx in range(4):
            sx: int = x + bx
            sy: int = y + by
            if sx < w and sy < h:
                idx: int = (sy * w + sx) * channels
                block.extend(src[idx : idx + channels])
                if channels == 3:
                    block.append(255)  # Add alpha for RGB
            else:
                block.extend([0] * 4)
    return block


def _compress_dxt1_block(
    dest: bytearray,
    src: list[int],
) -> None:
    _compress_color_block(dest, src)


def _compress_dxt3_block(
    dest: bytearray,
    src: list[int],
) -> None:
    _compress_alpha_block_dxt3(dest, src)
    _compress_color_block(dest[8:], src)


def _compress_alpha_block_dxt3(
    dest: bytearray,
    src: list[int],
) -> None:
    for i in range(8):
        alpha1: int = src[i * 8 + 3] >> 4
        alpha2: int = src[i * 8 + 7] >> 4
        dest[i] = (alpha2 << 4) | alpha1


def _compress_dxt5_block(dest: bytearray, src: list[int]) -> None:
    _compress_alpha_block_dxt5(dest, src)
    _compress_color_block(dest[8:], src)


def _compress_alpha_block_dxt5(
    dest: bytearray,
    src: list[int],
) -> None:
    alpha: list[int] = [src[i * 4 + 3] for i in range(16)]
    min_a: int = min(alpha)
    max_a: int = max(alpha)

    if min_a == max_a:
        dest[0] = max_a
        dest[1] = min_a
        dest[2:8] = [0] * 6
        return

    if min_a > max_a:
        min_a, max_a = max_a, min_a

    dest[0] = max_a
    dest[1] = min_a

    indices = 0
    for i in range(16):
        if alpha[i] == 0:
            code = 6
        elif alpha[i] == 255:  # noqa: PLR2004
            code = 7
        else:
            t: int = (alpha[i] - min_a) * 7 // (max_a - min_a)
            code = min(7, t)

        indices |= code << (3 * i)

    for i in range(6):
        dest[2 + i] = (indices >> (8 * i)) & 255


def _compress_color_block(
    dest: bytearray,
    src: list[int],
) -> None:
    if all(src[i : i + 4] == src[0:4] for i in range(4, 64, 4)):
        r, g, b = src[0], src[1], src[2]
        mask: int = 0xAAAAAAAA
        max16: int = as_16bit(quantize_rb(r), quantize_g(g), quantize_rb(b))
        min16: int = as_16bit(quantize_rb(r), quantize_g(g), quantize_rb(b))
    else:
        dblock: list[int] = dither_block(src)
        max16, min16 = optimize_colors_block(dblock)
        if max16 != min16:
            color = eval_colors(max16, min16)
            mask = match_colors_block(src, color)
        else:
            mask = 0

        for _ in range(2):
            lastmask: int = mask
            if refine_block(src, max16, min16, mask):
                if max16 != min16:
                    color = eval_colors(max16, min16)
                    mask = match_colors_block(src, color)
                else:
                    mask = 0
                    break
            if mask == lastmask:
                break

    if max16 < min16:
        max16, min16 = min16, max16
        mask ^= 0x55555555

    dest[0] = max16 & 0xFF
    dest[1] = max16 >> 8
    dest[2] = min16 & 0xFF
    dest[3] = min16 >> 8
    dest[4] = mask & 0xFF
    dest[5] = (mask >> 8) & 0xFF
    dest[6] = (mask >> 16) & 0xFF
    dest[7] = (mask >> 24) & 0xFF


def dither_block(block: list[int]) -> list[int]:
    dblock: list[int] = block.copy()
    err: list[int] = [0] * 8
    for ch in range(3):
        for y in range(4):
            for x in range(4):
                idx: int = (y * 4 + x) * 4 + ch
                old: int = dblock[idx]
                new: int = quantize_rb(old) if ch != 1 else quantize_g(old)
                dblock[idx] = new
                err_val: int = old - new
                if x < 3:  # noqa: PLR2004
                    dblock[idx + 4] += (err_val * 7) >> 4
                if y < 3:  # noqa: PLR2004
                    if x > 0:
                        dblock[idx + 12] += (err_val * 3) >> 4
                    dblock[idx + 16] += (err_val * 5) >> 4
                    if x < 3:
                        dblock[idx + 20] += err_val >> 4
    return dblock


def optimize_colors_block(block: list[int]) -> tuple[int, int]:
    cov: list[int] = [0] * 6
    mu: list[int] = [0] * 3
    min_color: list[int] = [255] * 3
    max_color: list[int] = [0] * 3

    for i in range(16):
        r, g, b = block[i * 4 : i * 4 + 3]
        mu[0] += r
        mu[1] += g
        mu[2] += b
        min_color[0] = min(min_color[0], r)
        min_color[1] = min(min_color[1], g)
        min_color[2] = min(min_color[2], b)
        max_color[0] = max(max_color[0], r)
        max_color[1] = max(max_color[1], g)
        max_color[2] = max(max_color[2], b)

    mu[0] = (mu[0] + 8) >> 4
    mu[1] = (mu[1] + 8) >> 4
    mu[2] = (mu[2] + 8) >> 4

    for i in range(16):
        r = block[i * 4] - mu[0]
        g = block[i * 4 + 1] - mu[1]
        b = block[i * 4 + 2] - mu[2]
        cov[0] += r * r
        cov[1] += r * g
        cov[2] += r * b
        cov[3] += g * g
        cov[4] += g * b
        cov[5] += b * b

    vfr: int = max_color[0] - min_color[0]
    vfg: int = max_color[1] - min_color[1]
    vfb: int = max_color[2] - min_color[2]

    for _ in range(4):  # Power iteration
        r = vfr * cov[0] + vfg * cov[1] + vfb * cov[2]
        g = vfr * cov[1] + vfg * cov[3] + vfb * cov[4]
        b = vfr * cov[2] + vfg * cov[4] + vfb * cov[5]
        vfr, vfg, vfb = r, g, b

    magn: int = max(abs(vfr), abs(vfg), abs(vfb))
    if magn < 4.0:  # noqa: PLR2004
        v_r, v_g, v_b = 299, 587, 114
    else:
        v_r = int(vfr * 512 / magn)
        v_g = int(vfg * 512 / magn)
        v_b = int(vfb * 512 / magn)

    min_d: float = float("inf")
    max_d: float = float("-inf")
    min_p: int = 0
    max_p: int = 0
    for i in range(16):
        dot: int = block[i * 4] * v_r + block[i * 4 + 1] * v_g + block[i * 4 + 2] * v_b
        if dot < min_d:
            min_d = dot
            min_p = i
        if dot > max_d:
            max_d = dot
            max_p = i

    return (
        as_16bit(block[max_p * 4], block[max_p * 4 + 1], block[max_p * 4 + 2]),
        as_16bit(block[min_p * 4], block[min_p * 4 + 1], block[min_p * 4 + 2]),
    )


def eval_colors(
    color0: int,
    color1: int,
) -> list[int]:
    def expand_565(c: int) -> tuple[int, int, int]:
        return ((c >> 11) & 31) * 8, ((c >> 5) & 63) * 4, (c & 31) * 8

    c0: tuple[int, int, int] = expand_565(color0)
    c1: tuple[int, int, int] = expand_565(color1)

    colors: list[int] = [
        c0[0],
        c0[1],
        c0[2],
        255,
        c1[0],
        c1[1],
        c1[2],
        255,
        (2 * c0[0] + c1[0]) // 3,
        (2 * c0[1] + c1[1]) // 3,
        (2 * c0[2] + c1[2]) // 3,
        255,
        (c0[0] + 2 * c1[0]) // 3,
        (c0[1] + 2 * c1[1]) // 3,
        (c0[2] + 2 * c1[2]) // 3,
        255,
    ]

    return colors


def match_colors_block(
    block: list[int],
    color: list[int],
) -> int:
    mask: int = 0
    dir_r: int = color[0] - color[4]
    dir_g: int = color[1] - color[5]
    dir_b: int = color[2] - color[6]
    dots: list[int] = [0] * 16
    stops: list[int] = [0] * 4

    for i in range(16):
        dots[i] = block[i * 4] * dir_r + block[i * 4 + 1] * dir_g + block[i * 4 + 2] * dir_b

    for i in range(4):
        stops[i] = color[i * 4] * dir_r + color[i * 4 + 1] * dir_g + color[i * 4 + 2] * dir_b

    c0_point: int = (stops[1] + stops[3]) >> 1
    half_point: int = (stops[3] + stops[2]) >> 1
    c3_point: int = (stops[2] + stops[0]) >> 1

    for i in range(16):
        dot: int = dots[i]
        if dot < half_point:
            mask |= (3 if dot < c0_point else 1) << (i * 2)
        else:
            mask |= (2 if dot < c3_point else 0) << (i * 2)

    return mask


def refine_block(
    block: list[int],
    max16: int,
    min16: int,
    mask: int,
) -> bool:
    old_min, old_max = min16, max16
    at1_r = at1_g = at1_b = 0
    at2_r = at2_g = at2_b = 0
    akku = 0

    for i in range(16):
        step: int = mask & 3
        w1: int = [3, 0, 2, 1][step]
        r, g, b = block[i * 4 : i * 4 + 3]
        akku += [0x090000, 0x000900, 0x040102, 0x010402][step]
        at1_r += w1 * r
        at1_g += w1 * g
        at1_b += w1 * b
        at2_r += r
        at2_g += g
        at2_b += b
        mask >>= 2

    at2_r = 3 * at2_r - at1_r
    at2_g = 3 * at2_g - at1_g
    at2_b = 3 * at2_b - at1_b

    xx: int = akku >> 16
    yy: int = (akku >> 8) & 255
    xy: int = akku & 255

    denom: int = xx * yy - xy * xy
    if denom == 0:
        return False

    f_rb: float = (3 * 31) / 255 / denom
    f_g: float = f_rb * 63 / 31

    max16 = (
        (sclamp((at1_r * yy - at2_r * xy) * f_rb + 0.5, 0, 31) << 11)
        | (sclamp((at1_g * yy - at2_g * xy) * f_g + 0.5, 0, 63) << 5)
        | sclamp((at1_b * yy - at2_b * xy) * f_rb + 0.5, 0, 31)
    )

    min16 = (
        (sclamp((at2_r * xx - at1_r * xy) * f_rb + 0.5, 0, 31) << 11)
        | (sclamp((at2_g * xx - at1_g * xy) * f_g + 0.5, 0, 63) << 5)
        | sclamp((at2_b * xx - at1_b * xy) * f_rb + 0.5, 0, 31)
    )

    return old_min != min16 or old_max != max16


def as_16bit(
    r: int,
    g: int,
    b: int,
) -> int:
    return ((r >> 3) << 11) | ((g >> 2) << 5) | (b >> 3)


def sclamp(
    y: float,
    p0: int,
    p1: int,
) -> int:
    x: int = int(y)
    return max(p0, min(x, p1))


def quantize_rb(x: int) -> int:
    return (x * 31 + 127) // 255


def quantize_g(x: int) -> int:
    return (x * 63 + 127) // 255



"""CSV read/write for 2DA; first column is row label, rest are data columns."""

from __future__ import annotations

import csv
import io

from typing import TYPE_CHECKING

from pykotor.resource.formats.twoda.twoda_data import TwoDA
from pykotor.resource.type import ResourceReader, ResourceWriter, autoclose
from pykotor.tools.encoding import decode_bytes_with_fallbacks

if TYPE_CHECKING:
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES


class TwoDACSVReader(ResourceReader):
    """Reads 2DA files from CSV format.

    CSV is a PyKotor-specific convenience format for easier editing in spreadsheet applications.
    Format: First column is row label, remaining columns are headers.

    References:
    ----------
        Based on swkotor.exe 2DA structure:
        - C2DA::Load2DArray @ 0x004143b0 - Loads 2DA file from resource
          * Parses "2DA V2.0" header
          * Handles "DEFAULT:" line for default cell values
          * Reads column headers (tab-separated)
          * Reads row labels and cell data
        - C2DA::Unload2DArray @ 0x004139e0 - Unloads 2DA data
        - " 2DA file" string @ 0x0074b328 - 2DA file identifier
        - Error messages for missing 2DA files:
          * "CSWClass::LoadFeatGain: can't load featgain.2da" @ 0x0074b370
          * "CSWClass::LoadFeatTable: Can't load feat.2da" @ 0x0074b3c8
          * "CSWClass::LoadSkillsTable: Can't load skills.2da" @ 0x0074b454
          * "CSWClass::LoadSpellsTable: Can't load spells.2da" @ 0x0074b5c0

        Note: CSV format is PyKotor-specific, not a standard game format.
        The engine uses binary 2DA format exclusively. CSV conversion allows easier editing
        in spreadsheet applications.
    """

    def __init__(
        self,
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = 0,
    ):
        super().__init__(source, offset, size)
        self._twoda: TwoDA | None = None

    @autoclose
    def load(self, *, auto_close: bool = True) -> TwoDA:  # noqa: FBT001, FBT002, ARG002
        self._twoda = TwoDA()
        data: str = decode_bytes_with_fallbacks(self._reader.read_bytes(self._reader.size()))
        _csv = csv.reader(io.StringIO(data))

        try:
            headers: list[str] = next(_csv)[1:]
            if not headers:
                msg = "CSV header is missing or not formatted correctly."
                raise ValueError(msg)

            for header in headers:
                if not header.strip():
                    msg = "Empty header detected, CSV is not valid."
                    raise ValueError(msg)
                self._twoda.add_column(header.strip())

            for i, row in enumerate(_csv, start=1):
                if len(row) != len(headers) + 1:
                    msg = f"Row {i} does not have the correct number of columns."
                    raise ValueError(msg)

                label = row[0]
                if not label.strip():
                    msg = f"Row {i} does not have a valid label."
                    raise ValueError(msg)

                cells = dict(zip(headers, row[1:]))
                self._twoda.add_row(label.strip(), cells)
        except csv.Error as e:
            msg = f"CSV reader error: {e!r}"
            raise ValueError(msg) from e

        return self._twoda


class TwoDACSVWriter(ResourceWriter):
    def __init__(
        self,
        twoda: TwoDA,
        target: TARGET_TYPES,
    ):
        super().__init__(target)
        self._twoda: TwoDA = twoda
        self._csv_string = io.StringIO("")
        self._csv_writer = csv.writer(self._csv_string)

    @autoclose
    def write(self, *, auto_close: bool = True):  # noqa: FBT001, FBT002, ARG002  # pyright: ignore[reportUnusedParameters]
        headers: list[str] = self._twoda.get_headers()

        insert: list[str] = [""]
        insert.extend(iter(headers))
        self._csv_writer.writerow(insert)

        for row in self._twoda:
            insert = [str(row.label())]
            insert.extend(row.get_string(header) for header in headers)
            self._csv_writer.writerow(insert)

        data: bytes = self._csv_string.getvalue().encode("ascii")
        self._writer.write_bytes(data)



"""Module abstraction: load/save ARE, GIT, LYT, VIS, and resources from capsule or path."""

from __future__ import annotations

import errno
import os

from contextlib import suppress
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Collection, Generic, TypeVar, TypedDict, cast

from loggerplus import RobustLogger
from pykotor.common.language import LocalizedString
from pykotor.common.misc import ResRef
from pykotor.extract.capsule import Capsule
from pykotor.extract.file import FileResource, LocationResult, ResourceIdentifier
from pykotor.extract.installation import SearchLocation
from pykotor.resource.formats.bwm.bwm_auto import bytes_bwm, read_bwm, write_bwm
from pykotor.resource.formats.erf.erf_auto import read_erf, write_erf
from pykotor.resource.formats.gff import read_gff
from pykotor.resource.formats.gff.gff_data import GFF, GFFFieldType
from pykotor.resource.formats.lyt.lyt_auto import bytes_lyt, read_lyt, write_lyt
from pykotor.resource.formats.lyt.lyt_data import LYT
from pykotor.resource.formats.ncs.ncs_auto import write_ncs
from pykotor.resource.formats.rim.rim_auto import read_rim, write_rim
from pykotor.resource.formats.tpc.tpc_auto import bytes_tpc, read_tpc, write_tpc
from pykotor.resource.formats.vis.vis_auto import bytes_vis, read_vis, write_vis
from pykotor.resource.formats.vis.vis_data import VIS
from pykotor.resource.generics.are import bytes_are, read_are, write_are
from pykotor.resource.generics.dlg import bytes_dlg, read_dlg, write_dlg
from pykotor.resource.generics.git import GIT, bytes_git, read_git, write_git
from pykotor.resource.generics.ifo import IFO, bytes_ifo, read_ifo, write_ifo
from pykotor.resource.generics.pth import bytes_pth, read_pth, write_pth
from pykotor.resource.generics.utc import UTC, bytes_utc, read_utc, write_utc
from pykotor.resource.generics.utd import UTD, bytes_utd, read_utd, write_utd
from pykotor.resource.generics.ute import UTE, bytes_ute, read_ute, write_ute
from pykotor.resource.generics.uti import bytes_uti, read_uti, write_uti
from pykotor.resource.generics.utm import UTM, bytes_utm, read_utm, write_utm
from pykotor.resource.generics.utp import UTP, bytes_utp, read_utp, write_utp
from pykotor.resource.generics.uts import UTS, bytes_uts, read_uts, write_uts
from pykotor.resource.generics.utt import UTT, bytes_utt, read_utt, write_utt
from pykotor.resource.generics.utw import UTW, bytes_utw, read_utw, write_utw
from pykotor.resource.type import ResourceType
from pykotor.tools.misc import is_any_erf_type_file, is_bif_file, is_capsule_file, is_rim_file
from pykotor.tools.model import iterate_lightmaps, iterate_textures
from pykotor.tools.path import CaseAwarePath

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Iterable

    from typing_extensions import Self  # pyright: ignore[reportMissingModuleSource]

    from pykotor.common.misc import Game
    from pykotor.extract.file import LocationResult, ResourceResult
    from pykotor.extract.installation import Installation
    from pykotor.resource.formats.erf.erf_data import ERF
    from pykotor.resource.formats.gff.gff_data import GFF, GFFList
    from pykotor.resource.formats.mdl.mdl_data import MDL
    from pykotor.resource.formats.rim.rim_data import RIM
    from pykotor.resource.formats.tpc.tpc_data import TPC
    from pykotor.resource.generics.are import ARE
    from pykotor.resource.generics.ifo import IFO
    from pykotor.resource.generics.pth import PTH
    from pykotor.resource.generics.uti import UTI
    from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES

T = TypeVar("T")
SEARCH_ORDER: list[SearchLocation] = [
    SearchLocation.OVERRIDE,
    SearchLocation.CUSTOM_MODULES,
    SearchLocation.CHITIN,
]


class KModuleType(Enum):
    """Module file type enumeration.

    KotOR modules are split across multiple archive files. The module system
    uses different file extensions to organize resources by type and priority.

    References:
    ----------
        Original BioWare Odyssey Engine (module archive structure from swkotor.exe, swkotor2.exe)
        Note: Module file organization varies between KotOR 1 and KotOR 2


        Note: Module file organization varies between KotOR 1 and KotOR 2

    File Organization:
    -----------------
        - MAIN (.rim): Contains core module files (IFO, ARE, GIT) - only loaded in simple mode
        - AREA (_a.rim): Area-specific RIM (complex mode) - replaces .rim
        - AREA_EXTENDED (_adx.rim): Extended area RIM (complex mode) - replaces .rim if _a.rim not found
        - DATA (_s.rim): Contains module data (creatures, items, placeables, etc.)
        - K2_DLG (_dlg.erf): KotOR 2 only - contains dialog files
        - MOD (.mod): Community override format, replaces all above files

    Reverse Engineering Notes:
    -------------------------
        Based on swkotor.exe: FUN_004094a0 and swkotor2.exe: FUN_004096b0
        - Simple Mode (flag at offset 0x54 == 0): Loads .rim directly
        - Complex Mode (flag != 0): Checks for _a.rim, _adx.rim, .mod, _s.rim, _dlg.erf
        - _a.rim and _adx.rim REPLACE .rim in complex mode
        - .mod REPLACES all other files if it exists
    """

    MAIN = ".rim"  # Contains the IFO/ARE/GIT
    """Main module archive containing core module files.
    
    Reference: Original BioWare Odyssey Engine module structure
    Contains: IFO (module info), ARE (area data), GIT (dynamic area info)
    File naming: <modulename>.rim
    Note: Only loaded in simple mode (flag at offset 0x54 == 0)
    swkotor.exe: FUN_004094a0 line 32-42
    """

    AREA = "_a.rim"  # Area-specific RIM (complex mode)
    """Area-specific module archive (complex mode).
    
    Reference: swkotor.exe: FUN_004094a0 line 61, 159
    Contains: ARE (area data) and related area resources
    File naming: <modulename>_a.rim
    Behavior: REPLACES .rim in complex mode
    """

    AREA_EXTENDED = "_adx.rim"  # Extended area RIM (complex mode)
    """Extended area module archive (complex mode).
    
    Reference: swkotor.exe: FUN_004094a0 line 74, 85
    Contains: ARE (area data) and extended area resources
    File naming: <modulename>_adx.rim
    Behavior: REPLACES .rim if _a.rim not found
    """

    DATA = "_s.rim"  # Contains everything else
    """Data module archive containing module resources.
    
    Reference: Original BioWare Odyssey Engine module structure
    Contains: UTC, UTD, UTE, UTI, UTM, UTP, UTS, UTT, UTW, FAC, LYT, NCS, PTH
    File naming: <modulename>_s.rim
    Note: In KotOR 2, DLG files are NOT in _s.rim (see K2_DLG)
    """

    K2_DLG = "_dlg.erf"  # In TSL, DLGs are here instead of _s.rim.
    """KotOR 2 dialog archive containing dialog files.
    
    Reference: Original BioWare Odyssey Engine (KotOR 2 only)
    Contains: DLG (dialog) files
    File naming: <modulename>_dlg.erf
    Note: KotOR 1 stores DLG files in _s.rim, KotOR 2 uses separate _dlg.erf
    """

    MOD = ".mod"  # Community-standard override, takes priority over the above 3 files. This extension overrides all 3 of the above, while the other 3 are complementary to each other.
    """Community override module archive (single-file format).
    
    Reference: TSLPatcher modding community standard
    Contains: All module resources in a single ERF archive
    File naming: <modulename>.mod
    Priority: Takes precedence over .rim/_s.rim/_dlg.erf files
    Note: This is a modding convention, not used by the original game engine
    """

    def contains(  # noqa: PLR0911
        self,
        restype: ResourceType,
        *,
        game: Game | None = None,
    ) -> bool:
        """Whether this ModuleType, if not modified, would contain the specified ResourceType or not.

        Args:
        ----
            restype (ResourceType): The type of resource to check.
            game (Game | None = None): Optional game to determine whether to invalidate _s.rim if they contain DLGs.

        Returns:
        -------
            bool: Whether this ModuleType should contain the ResourceType or not.
        """
        if restype.target_type() is not restype:
            return False
        if restype == ResourceType.DLG:
            if game is None:
                return self is self.DATA or self is self.K2_DLG
            if game.is_k1():
                return self is self.DATA
            if game.is_k2():
                return self is self.K2_DLG

        if self is self.MOD:
            return self is not ResourceType.TwoDA
        if self is self.MAIN:
            return restype in {ResourceType.ARE, ResourceType.IFO, ResourceType.GIT}
        # _a.rim and _adx.rim contain ARE resources (swkotor.exe: FUN_004094a0 line 61, 74 - checks for ARE type 0xbba)
        if self is self.AREA or self is self.AREA_EXTENDED:
            return restype in {ResourceType.ARE, ResourceType.IFO, ResourceType.GIT}
        if self is self.DATA:
            return restype in {
                ResourceType.FAC,
                ResourceType.LYT,
                ResourceType.NCS,
                ResourceType.PTH,
                ResourceType.UTC,
                ResourceType.UTD,
                ResourceType.UTE,
                ResourceType.UTI,
                ResourceType.UTM,
                ResourceType.UTP,
                ResourceType.UTS,
                ResourceType.UTT,
                ResourceType.UTW,
            }
        raise RuntimeError(f"Invalid ModuleType enum: {self!r}")


@dataclass(frozen=True)
class ModulePieceInfo:
    root: str
    modtype: KModuleType

    @classmethod
    def from_filename(
        cls,
        filename: str | ResourceIdentifier,
    ) -> Self:
        if isinstance(filename, ResourceIdentifier):
            filename = str(filename)
        root = Module.name_to_root(filename)
        return cls(root, KModuleType(filename[len(root) :]))

    def filename(self) -> str:
        return self.root + self.modtype.value

    def res_ident(self) -> ResourceIdentifier:
        return ResourceIdentifier.from_path(self.filename())

    def resname(self) -> str:
        """Resource name without extension; same as root for module pieces."""
        return self.root

    def restype(self) -> ResourceType:
        return ResourceType.from_extension(self.modtype.value)


class ModulePieceResource(Capsule):
    def __new__(
        cls,
        path: os.PathLike | str,
        *args,
        **kwargs,
    ):
        new_cls = cls
        if new_cls is ModulePieceResource:
            path_obj = CaseAwarePath(path)
            piece_info = ModulePieceInfo.from_filename(path_obj.name)
            if piece_info.modtype is KModuleType.DATA:
                new_cls = ModuleDataPiece
            elif piece_info.modtype is KModuleType.MAIN:
                new_cls = ModuleLinkPiece
            elif piece_info.modtype is KModuleType.AREA:
                new_cls = ModuleLinkPiece  # _a.rim uses same structure as .rim
            elif piece_info.modtype is KModuleType.AREA_EXTENDED:
                new_cls = ModuleLinkPiece  # _adx.rim uses same structure as .rim
            elif piece_info.modtype is KModuleType.K2_DLG:
                new_cls = ModuleDLGPiece
            elif piece_info.modtype is KModuleType.MOD:
                new_cls = ModuleFullOverridePiece
        return object.__new__(new_cls)  # pyright: ignore[reportArgumentType]

    def __init__(
        self,
        path: os.PathLike | str,
        *args,
        **kwargs,
    ):
        path_obj = CaseAwarePath(path)
        self.piece_info: ModulePieceInfo = ModulePieceInfo.from_filename(path_obj.name)
        self.missing_resources: list[FileResource] = []  # TODO(th3w1zard1):
        super().__init__(path_obj, *args, **kwargs)

    def __reduce__(self):
        # Return a tuple with the callable, arguments, and state
        return (self.__class__, (str(self.filepath()),), self.__getstate__())

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove or modify any non-picklable attributes if necessary
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)


class ModuleLinkPiece(ModulePieceResource):
    def ifo(self) -> GFF:
        lookup = self.resource("module", ResourceType.IFO)
        if lookup is None:
            import errno

            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(self.filepath().joinpath("module.ifo")))
        return read_gff(lookup)

    def module_id(self) -> ResRef | None:
        """Get the module id, attempt to just check resrefs, fallback to the Mod_Area_list."""
        link_resources: set[FileResource] = {
            resource for resource in self._resources if resource.restype() is not ResourceType.IFO and KModuleType.MAIN.contains(resource.restype())
        }
        if link_resources:
            check_resname = next(iter(link_resources)).identifier().lower_resname
            if all(check_resname == res.identifier().lower_resname for res in link_resources):
                print(f"Module ID, Check 1: All link resources have the same resref of '{check_resname}'")
                return ResRef(check_resname)

        gff_ifo: GFF = self.ifo()
        if gff_ifo.root.exists("Mod_Area_list"):
            actual_ftype: GFFFieldType = gff_ifo.root.what_type("Mod_Area_list")
            if actual_ftype != GFFFieldType.List:
                RobustLogger().warning(f"{self.filename()} has IFO with incorrect field 'Mod_Area_list' type '{actual_ftype.name}', expected 'List'")
            else:
                area_list: GFFList | None = gff_ifo.root.get_list("Mod_Area_list")
                if area_list is None:
                    RobustLogger().error(f"{self.filename()}: Module.IFO has a Mod_Area_list field, but it is not a valid list.")
                    return None
                area_localized_name: ResRef | None = next((gff_struct.get_resref("Area_Name") for gff_struct in area_list if gff_struct.exists("Area_Name")), None)
                if area_localized_name is not None and str(area_localized_name).strip():
                    print(f"Module ID, Check 2: Found in Mod_Area_list: '{area_localized_name}'")
                    return area_localized_name
            print(f"{self.filename()}: Module.IFO does not contain a valid Mod_Area_list. Could not get the module id!")
        else:
            RobustLogger().error(f"{self.filename()}: Module.IFO does not have an existing Mod_Area_list.")
        return None

    def area_name(self) -> LocalizedString | ResRef:
        """See if the ARE is already cached, otherwise use the fallback."""
        area_file_res: FileResource | None = next((resource for resource in self._resources if resource.restype() == ResourceType.ARE), None)
        if area_file_res is not None:
            gff_are: GFF = read_gff(area_file_res.data())
            if gff_are.root.exists("Name"):
                actual_ftype: GFFFieldType = gff_are.root.what_type("Name")
                if actual_ftype != GFFFieldType.LocalizedString:
                    raise ValueError(f"{self.filename()} has IFO with incorrect field 'Name' type '{actual_ftype.name}', expected 'LocalizedString'")
                result: LocalizedString | None = gff_are.root.get_locstring("Name")
                if result is None:
                    RobustLogger().error(f"{self.filename()}: ARE has a Name field, but it is not a valid LocalizedString.")
                    return LocalizedString.from_invalid()
                print(f"Check 1 result: '{result}'")
                return result
        raise ValueError(f"Failed to find an ARE for module '{self.piece_info.filename()}'")


class ModuleDataPiece(ModulePieceResource): ...


class ModuleDLGPiece(ModulePieceResource): ...


class ModuleFullOverridePiece(ModuleDLGPiece, ModuleDataPiece, ModuleLinkPiece): ...


class _CapsuleDictTypes(TypedDict, total=False):
    MAIN: ModuleLinkPiece | None
    AREA: ModuleLinkPiece | None
    AREA_EXTENDED: ModuleLinkPiece | None
    DATA: ModuleDataPiece | None
    K2_DLG: ModuleDLGPiece | None
    MOD: ModuleFullOverridePiece | None


class Module:  # noqa: PLR0904
    """Represents a KotOR game module with its resources and archives.

    A Module aggregates resources from multiple archive files (.rim, _s.rim, _dlg.erf)
    or a single override archive (.mod). It manages resource loading, activation,
    and provides access to module-specific resources like areas, creatures, items, etc.

    References:
    ----------
        Based on swkotor.exe module system:
        - CServerExoAppInternal::LoadModule @ 0x004b95b0 - Main module loader (4088 bytes, 36 callees)
          * Loads module from IFO file
          * Handles module initialization and resource loading
          * Calls LoadModuleStart, LoadModuleInProgress, LoadModuleFinish
        - CSWSModule::LoadModuleStart @ 0x004c9050 - Module start loader (5666 bytes, 68 callees)
          * Initializes module resources
          * Loads IFO, ARE, GIT files
          * Sets up module area and objects
        - LoadModuleInProgress (/K1/k1_win_gog_swkotor.exe @ 0x004c5720, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module loading progress handler (347 bytes, 7 callees)
        - LoadModuleFinish (/K1/k1_win_gog_swkotor.exe @ 0x004c5880, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module loading completion handler (245 bytes, 9 callees)
        - UnloadModule (/K1/k1_win_gog_swkotor.exe @ 0x004b9240, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Unloads module (850 bytes, 32 callees)
        - FUN_004094a0 - Module archive loader (loads .rim, _s.rim, _adx.rim files)
          * Handles module archive file discovery
          * Loads main .rim, data _s.rim, and extended _adx.rim archives
        - "MODULES:" string (/K1/k1_win_gog_swkotor.exe @ 0x0073d90c, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module list identifier
        - "Module" string (/K1/k1_win_gog_swkotor.exe @ 0x007442a8, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module identifier
        - "CSWSSCRIPTEVENT_EVENTTYPE_ON_MODULE_LOAD" (/K1/k1_win_gog_swkotor.exe @ 0x007446e4, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module load event
        - "CSWSSCRIPTEVENT_EVENTTYPE_ON_MODULE_START" (/K1/k1_win_gog_swkotor.exe @ 0x00744710, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module start event
        - "ModuleList" (/K1/k1_win_gog_swkotor.exe @ 0x00745044, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address), "GetModuleList" (/K1/k1_win_gog_swkotor.exe @ 0x00745050, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module list functions
        - "ModuleRunning" (/K1/k1_win_gog_swkotor.exe @ 0x00745060, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address), "ModuleLoaded" (/K1/k1_win_gog_swkotor.exe @ 0x00745078, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module state strings
        - "modulesave" (/K1/k1_win_gog_swkotor.exe @ 0x00745128, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address), "ModuleName" (/K1/k1_win_gog_swkotor.exe @ 0x00745134, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Module save/name strings

    Attributes:
    ----------
        resources: Dictionary mapping ResourceIdentifier to ModuleResource.
            All resources available in this module, keyed by identifier for uniqueness.

        dot_mod: Whether this module uses .mod override format.
            Reference: TSLPatcher modding convention
            If True, uses <root>.mod archive; if False, uses .rim/_s.rim/_dlg.erf archives.

        _installation: Cached Installation instance for resource lookups.
            Used to resolve resources from chitin, override, and other locations.

        _root: Root module name (without extensions).
            Extracted from filename, used to construct archive filenames.

        _cached_mod_id: Cached module ResRef identifier.
            Module identifier extracted from IFO or archive filenames.

        _cached_sort_id: Cached sort identifier for module ordering.
            PyKotor-specific: Used for module sorting/ordering in tools.

        _capsules: Dictionary of module archive capsules.
            Contains ModuleLinkPiece, ModuleDataPiece, ModuleDLGPiece, or ModuleFullOverridePiece
            depending on module type and available files.
    """

    def __init__(
        self,
        filename_or_root: str,  # The root name of the module.
        installation: Installation,  # Cached installation instance.
        *,
        use_dot_mod: bool = True,  # Should this Module instance represent the .rim/_s.rim/._dlg.erf vanilla archives, or the singular `root`.mod override archive?
        load_textures: bool = True,  # Whether to crawl model references to locate textures/lightmaps.
    ):
        self.resources: dict[ResourceIdentifier, ModuleResource] = {}  # The keys are only used for ensured uniqueness.
        self.dot_mod: bool = use_dot_mod
        self._installation: Installation = installation
        self._root: str = self.name_to_root(filename_or_root.lower())
        self._cached_mod_id: ResRef | None = None
        self._cached_sort_id: str | None = None
        self._load_textures: bool = load_textures

        # Build all capsules relevant to this root in the provided installation
        # Based on swkotor.exe: FUN_004094a0 and swkotor2.exe: FUN_004096b0
        self._capsules: _CapsuleDictTypes = {
            KModuleType.MAIN.name: None,
            KModuleType.AREA.name: None,
            KModuleType.AREA_EXTENDED.name: None,
            KModuleType.DATA.name: None,
            KModuleType.K2_DLG.name: None,
            KModuleType.MOD.name: None,
        }
        module_path = installation.module_path()

        if self.dot_mod:
            # .mod file overrides all rim-like files
            # swkotor.exe: FUN_004094a0 line 136: Loads .mod, skips _s.rim
            mod_filepath = module_path.joinpath(self._root + KModuleType.MOD.value)
            if mod_filepath.is_file():
                self._capsules[KModuleType.MOD.name] = ModuleFullOverridePiece(mod_filepath)  # pyright: ignore[reportGeneralTypeIssues]
            else:
                self.dot_mod = False
                # Fallback to rim files when .mod doesn't exist
                self._capsules[KModuleType.MAIN.name] = ModuleLinkPiece(module_path.joinpath(self._root + KModuleType.MAIN.value))  # pyright: ignore[reportGeneralTypeIssues]
                self._capsules[KModuleType.DATA.name] = ModuleDataPiece(module_path.joinpath(self._root + KModuleType.DATA.value))  # pyright: ignore[reportGeneralTypeIssues]
                if self._installation.game().is_k2():
                    self._capsules[KModuleType.K2_DLG.name] = ModuleDLGPiece(module_path.joinpath(self._root + KModuleType.K2_DLG.value))  # pyright: ignore[reportGeneralTypeIssues]
        else:
            # Complex mode: Check for _a.rim or _adx.rim (replaces .rim), then _s.rim and _dlg.erf
            # swkotor.exe: FUN_004094a0 line 49-216
            area_rim_path = module_path.joinpath(self._root + KModuleType.AREA.value)
            area_extended_rim_path = module_path.joinpath(self._root + KModuleType.AREA_EXTENDED.value)

            # Step 1: Load _a.rim if exists (REPLACES .rim)
            # swkotor.exe: FUN_004094a0 line 159
            if area_rim_path.is_file():
                self._capsules[KModuleType.AREA.name] = ModuleLinkPiece(area_rim_path)  # pyright: ignore[reportGeneralTypeIssues]
            # Step 2: Load _adx.rim if _a.rim not found (REPLACES .rim)
            # swkotor.exe: FUN_004094a0 line 85
            elif area_extended_rim_path.is_file():
                self._capsules[KModuleType.AREA_EXTENDED.name] = ModuleLinkPiece(area_extended_rim_path)  # pyright: ignore[reportGeneralTypeIssues]
            else:
                # Simple mode: Just load .rim file directly
                # swkotor.exe: FUN_004094a0 line 32-42
                self._capsules[KModuleType.MAIN.name] = ModuleLinkPiece(module_path.joinpath(self._root + KModuleType.MAIN.value))  # pyright: ignore[reportGeneralTypeIssues]

            # Step 3: Load _s.rim if exists (ADDS to base)
            # swkotor.exe: FUN_004094a0 line 118 (only if .mod not found)
            data_rim_path = module_path.joinpath(self._root + KModuleType.DATA.value)
            if data_rim_path.is_file():
                self._capsules[KModuleType.DATA.name] = ModuleDataPiece(data_rim_path)  # pyright: ignore[reportGeneralTypeIssues]

            # Step 4: Load _dlg.erf if exists (K2 only, ADDS to base)
            # swkotor2.exe: FUN_004096b0 line 147 (only if .mod not found)
            if self._installation.game().is_k2():
                dlg_erf_path = module_path.joinpath(self._root + KModuleType.K2_DLG.value)
                if dlg_erf_path.is_file():
                    self._capsules[KModuleType.K2_DLG.name] = ModuleDLGPiece(dlg_erf_path)  # pyright: ignore[reportGeneralTypeIssues]

        self.reload_resources()

    @classmethod
    def get_capsules_dict_matching(
        cls,
        install_or_path: Installation | Path,
        filename: str,
    ) -> _CapsuleDictTypes:
        from pykotor.extract.installation import Installation

        root = cls.name_to_root(filename)
        # Build all capsules relevant to this root in the provided installation
        # Based on swkotor.exe: FUN_004094a0 and swkotor2.exe: FUN_004096b0
        capsules: _CapsuleDictTypes = {
            KModuleType.MAIN.name: None,
            KModuleType.AREA.name: None,
            KModuleType.AREA_EXTENDED.name: None,
            KModuleType.DATA.name: None,
            KModuleType.K2_DLG.name: None,
            KModuleType.MOD.name: None,
        }
        module_path: Path = install_or_path if isinstance(install_or_path, Path) else install_or_path.module_path()

        if filename.lower().endswith(".mod"):
            mod_filepath = module_path.joinpath(root + KModuleType.MOD.value)
            if mod_filepath.is_file():
                capsules[KModuleType.MOD.name] = ModuleFullOverridePiece(mod_filepath)  # pyright: ignore[reportGeneralTypeIssues]
            else:
                # Engine is permissive: fall back to rim files when .mod doesn't exist
                capsules[KModuleType.MAIN.name] = ModuleLinkPiece(module_path.joinpath(root + KModuleType.MAIN.value))  # pyright: ignore[reportGeneralTypeIssues]
                capsules[KModuleType.DATA.name] = ModuleDataPiece(module_path.joinpath(root + KModuleType.DATA.value))  # pyright: ignore[reportGeneralTypeIssues]
                if not isinstance(install_or_path, Installation) or install_or_path.game().is_k2():
                    capsules[KModuleType.K2_DLG.name] = ModuleDLGPiece(module_path.joinpath(root + KModuleType.K2_DLG.value))  # pyright: ignore[reportGeneralTypeIssues]
        else:
            # Complex mode: Check for _a.rim or _adx.rim (replaces .rim), then _s.rim and _dlg.erf
            area_rim_path = module_path.joinpath(root + KModuleType.AREA.value)
            area_extended_rim_path = module_path.joinpath(root + KModuleType.AREA_EXTENDED.value)

            # Load _a.rim if exists (REPLACES .rim)
            if area_rim_path.is_file():
                capsules[KModuleType.AREA.name] = ModuleLinkPiece(area_rim_path)  # pyright: ignore[reportGeneralTypeIssues]
            # Load _adx.rim if _a.rim not found (REPLACES .rim)
            elif area_extended_rim_path.is_file():
                capsules[KModuleType.AREA_EXTENDED.name] = ModuleLinkPiece(area_extended_rim_path)  # pyright: ignore[reportGeneralTypeIssues]
            else:
                # Simple mode: Just .rim file
                capsules[KModuleType.MAIN.name] = ModuleLinkPiece(module_path.joinpath(root + KModuleType.MAIN.value))  # pyright: ignore[reportGeneralTypeIssues]

            # Load _s.rim if exists (ADDS to base)
            data_rim_path = module_path.joinpath(root + KModuleType.DATA.value)
            if data_rim_path.is_file():
                capsules[KModuleType.DATA.name] = ModuleDataPiece(data_rim_path)  # pyright: ignore[reportGeneralTypeIssues]

            # Load _dlg.erf if exists (K2 only, ADDS to base)
            if not isinstance(install_or_path, Installation) or install_or_path.game().is_k2():
                dlg_erf_path = module_path.joinpath(root + KModuleType.K2_DLG.value)
                if dlg_erf_path.is_file():
                    capsules[KModuleType.K2_DLG.name] = ModuleDLGPiece(dlg_erf_path)  # pyright: ignore[reportGeneralTypeIssues, reportUnknownReturnType]
        return capsules

    @classmethod
    def get_capsules_tuple_matching(
        cls,
        install_or_path: Installation | Path,
        filename: str,
    ) -> tuple[Capsule, ...]:
        capsules: _CapsuleDictTypes = cls.get_capsules_dict_matching(install_or_path, filename)
        return tuple(capsule for capsule in capsules.values() if isinstance(capsule, Capsule))

    def get_capsules(self) -> list[ModulePieceResource]:
        """Returns all relevant ERFs/RIMs for this module."""
        return list(self._capsules.values())  # pyright: ignore[reportReturnType]  # type: ignore[arg-type]

    def root(self) -> str:
        return self._root.strip()

    def lookup_main_capsule(self) -> ModuleFullOverridePiece | ModuleLinkPiece:
        """Returns main capsule either from the override or the module."""
        relevant_capsule: ModuleFullOverridePiece | ModuleLinkPiece | None
        if self.dot_mod:
            # .mod overrides all
            if KModuleType.MOD.name in self._capsules:
                relevant_capsule = self._capsules[KModuleType.MOD.name]
            else:
                relevant_capsule = self._capsules[KModuleType.MAIN.name]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        else:
            # Complex mode: Check for _a.rim or _adx.rim first (replaces .rim)
            # Simple mode: Just .rim
            if KModuleType.AREA.name in self._capsules and self._capsules[KModuleType.AREA.name] is not None:
                relevant_capsule = self._capsules[KModuleType.AREA.name]
            elif KModuleType.AREA_EXTENDED.name in self._capsules and self._capsules[KModuleType.AREA_EXTENDED.name] is not None:
                relevant_capsule = self._capsules[KModuleType.AREA_EXTENDED.name]
            else:
                relevant_capsule = self._capsules[KModuleType.MAIN.name]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        assert relevant_capsule is not None
        return relevant_capsule

    def lookup_data_capsule(self) -> ModuleFullOverridePiece | ModuleDataPiece:
        """Returns data capsule either from the override or the module."""
        relevant_capsule: ModuleFullOverridePiece | ModuleDataPiece | None
        if self.dot_mod:
            if KModuleType.MOD.name in self._capsules:
                relevant_capsule = self._capsules[KModuleType.MOD.name]
            else:
                relevant_capsule = self._capsules[KModuleType.DATA.name]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        else:
            relevant_capsule = self._capsules[KModuleType.DATA.name]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        assert relevant_capsule is not None
        return relevant_capsule

    def lookup_dlg_capsule(self) -> ModuleFullOverridePiece | ModuleDLGPiece:
        """Returns dlg capsule either from the override or the module."""
        relevant_capsule: ModuleFullOverridePiece | ModuleDLGPiece | None
        if self.dot_mod:
            if KModuleType.MOD.name in self._capsules:
                relevant_capsule = self._capsules[KModuleType.MOD.name]
            else:
                relevant_capsule = self._capsules[KModuleType.K2_DLG.name]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        else:
            relevant_capsule = self._capsules[KModuleType.K2_DLG.name]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        assert relevant_capsule is not None
        return relevant_capsule

    def module_id(self) -> ResRef | None:
        """Returns the module id from the main capsule."""
        if self._cached_mod_id is not None:
            return self._cached_mod_id
        data_capsule: ModuleFullOverridePiece | ModuleLinkPiece = self.lookup_main_capsule()
        found_id: ResRef | None = data_capsule.module_id()
        print(f"Found module id '{found_id}' for module '{data_capsule.filename()}'")
        self._cached_mod_id = found_id
        return found_id

    @staticmethod
    @lru_cache(maxsize=5000)
    def name_to_root(name: str) -> str:
        """Extracts the root module name from a string path or filename.

        This method strips any path components, file extensions, and common module suffixes
        (_s, _dlg) to get the base module name. The result is cached for performance.

        Args:
        ----
            name: A string containing a path or filename to extract the module name from.
                 Can be a full path or just a filename.

        Returns:
        -------
            str: The root module name with suffixes and extensions removed.
                 For example "danm13_s.rim" becomes "danm13".

        Example:
        -------
            >>> Module.name_to_root("c:/games/kotor/modules/danm13_s.rim")
            'danm13'
            >>> Module.name_to_root("danm13_dlg.dlg")
            'danm13'
        """
        split_path: list[str] = name.rsplit("/", 1)
        parsed_name = split_path[-1]
        name_without_ext: str = parsed_name.rsplit(".", 1)[0]
        root: str = name_without_ext.strip()
        casefold_root: str = root.casefold()
        root = root[:-2] if casefold_root.endswith("_s") else root
        root = root[:-4] if casefold_root.endswith("_dlg") else root
        root = root[:-2] if casefold_root.endswith("_a") else root
        root = root[:-4] if casefold_root.endswith("_adx") else root
        return root  # noqa: RET504

    @staticmethod
    def filepath_to_root(filepath: os.PathLike | str) -> str:
        """Converts a filesystem path to a module root name.

        This is a convenience wrapper around name_to_root() that handles PathLike objects.
        It converts the path to a string before extracting the module name.

        Args:
        ----
            filepath: A path-like object pointing to a module file.
                     Can be a `pathlib.Path`, `str`, or other path-like object.

        Returns:
        -------
            str: The root module name with suffixes and extensions removed.
                 For example `pathlib.Path("danm13_s.rim")` becomes `"danm13"`.

        See Also:
        --------
            name_to_root: The underlying implementation that does the actual conversion.
        """
        assert not isinstance(filepath, bytes)
        filepath_str: str = os.fspath(filepath)
        return Module.name_to_root(filepath_str)

    def capsules(self) -> list[ModulePieceResource]:
        """Returns a copy of the capsules used by the module.

        Returns:
        -------
            A tuple of linked capsules.
        """
        return [cast("ModulePieceResource", cap) for cap in self._capsules.values() if cap is not None]

    def reload_resources(self):
        """Reloads and updates the module's resources.

        Responsible for loading and updating the resources associated with the module.
        It processes the resources from various locations such as the module's own files, override directories,
        and core resources. The method also handles the identification and activation of resources linked
        through GIT, LYT, and VIS files.

        Processing Logic:
        ----------------
        1. Determine the display name of the module based on its type (.mod or.rim).
        2. Log the start of the resource loading process.
        3. Identify the main capsule to search for resources.
        4. Define the search order for resources (OVERRIDE, CUSTOM_MODULES, CHITIN).
        5. Create queries for GIT, LYT, and VIS resources using the module ID.
        6. Iterate through the module's capsules to add initial resource locations from the module files.
        7. Find references to resources in GIT, LYT, and VIS files:
            a. Process each query (GIT, LYT, VIS) to add locations and collect resource identifiers
            b. Track all resources referenced by GIT, LYT, VIS
            c. Add locations for all found resources
        8. Process core resources and override directories to add and activate resource locations:
            a. Check chitin/core resources first
            b. Prioritize override resources by checking/activating last
        9. Identify and activate texture resources linked in models:
            a. Iterate through all models in the module
            b. Find textures and lightmaps referenced in each model
            c. Create search queries for TPC and TGA textures
            d. Add locations for found texture resources
        10. Iterate through all resources to ensure all are activated, skipping duplicates:
            a. Skip TPC resources if the TGA equivalent is already found and activated
            b. Skip TGA resources if the TPC equivalent is already found and activated

        Raises:
        ------
            FileNotFoundError: If a required resource is not found in the expected locations.
            RuntimeError: If a resource type is unexpectedly None.
        """
        display_name = f"{self._root}.mod" if self.dot_mod else f"{self._root}.rim"
        RobustLogger().info("Loading module resources needed for '%s'", display_name)
        capsules_to_search: list[ModuleFullOverridePiece | ModuleLinkPiece] = [self.lookup_main_capsule()]
        # Lookup the GIT and LYT first.
        order: tuple[SearchLocation, ...] = (
            SearchLocation.OVERRIDE,
            SearchLocation.CUSTOM_MODULES,
            SearchLocation.CHITIN,
        )
        link_resname: str = str(self.module_id())
        lyt_query: ResourceIdentifier = ResourceIdentifier(link_resname, ResourceType.LYT)
        git_query: ResourceIdentifier = ResourceIdentifier(link_resname, ResourceType.GIT)
        vis_query: ResourceIdentifier = ResourceIdentifier(link_resname, ResourceType.VIS)

        # Start in our module resources, needs to happen first so we can determine what resources are part of our module.
        for capsule in self._capsules.values():
            if capsule is None:
                continue
            # No idea why static types aren't working here as that's the whole point of the TypedDict...
            typed_capsule: ModulePieceResource = cast("ModulePieceResource", capsule)
            for resource in typed_capsule:
                # RobustLogger().debug("Adding location '%s' for resource '%s' from erf/rim '%s'", typed_capsule.filepath(), resource.identifier(), typed_capsule.identifier())
                self.add_locations(resource.resname(), resource.restype(), [typed_capsule.filepath()])

        # Any resource referenced by the GIT/LYT/VIS not present in the module files
        # To be looked up elsewhere in the installation.
        main_search_results: dict[ResourceIdentifier, list[LocationResult]] = self._installation.locations(
            [lyt_query, git_query, vis_query],
            list(order),
            capsules=capsules_to_search,
        )

        # Track all resources referenced by GIT/LYT/VIS
        git_search: set[ResourceIdentifier] = set()
        lyt_search: set[ResourceIdentifier] = set()
        vis_search: set[ResourceIdentifier] = set()

        # Process each query (GIT, LYT, VIS) in sequence
        loop_iter_data: list[tuple[ResourceIdentifier, type[GIT | LYT | VIS], set[ResourceIdentifier]]] = [
            (git_query, GIT, git_search),
            (lyt_query, LYT, lyt_search),
            (vis_query, VIS, vis_search),
        ]
        for query, useable_type, search_set in loop_iter_data:
            if not main_search_results.get(query):
                if useable_type is VIS:
                    continue  # VIS is optional
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self.lookup_main_capsule().filepath() / str(query))

            # Add locations and get the resource wrapper
            resource_wrapper = self.add_locations(
                query.resname,
                query.restype,
                (loc.filepath for loc in main_search_results[query]),
            )
            # Activate each GIT/LYT location for this module, and fill this module with all of their resources (all of the resources their instances point to).
            # Store original path to restore later
            original_path: Path = resource_wrapper.locations()[0]

            # Check each location for referenced resources
            for location in resource_wrapper.locations():
                resource_wrapper.activate(location)
                loaded_resource: type[GIT | LYT | VIS] | None = resource_wrapper.resource()

                # VIS is optional and some installations ship malformed VIS. If parsing fails, treat as absent.
                if useable_type is VIS and loaded_resource is None:
                    continue
                if not isinstance(loaded_resource, (GIT, LYT, VIS)):
                    raise RuntimeError(f"{useable_type.__name__} is somehow None even though we know the path there.")  # noqa: TRY004

                # Only GIT/LYT have resource identifiers to collect
                if not isinstance(loaded_resource, VIS):
                    search_set.update(loaded_resource.iter_resource_identifiers())

            resource_wrapper.activate(original_path)

        # From GIT/LYT references, find them in the installation.
        search_results: dict[ResourceIdentifier, list[LocationResult]] = self._installation.locations(
            list({*git_search, *lyt_search, *vis_search}),
            list(order),
            capsules=capsules_to_search,
        )
        # Add locations for all found resources
        for identifier, locations in search_results.items():
            search_result_filepaths: tuple[Path, ...] = tuple(location.filepath for location in locations)
            self.add_locations(
                identifier.resname,
                identifier.restype,
                search_result_filepaths,
            )

        # Third. Since we now have a known full list of resources that make up this module, we can now process Override and chitin in one fell swoop.
        # Realistically we'll do this at the end, but right now we're interested in enumerating the models so we can find textures.
        # Check chitin first.
        for resource in self._installation.core_resources():
            if resource.identifier() in self.resources or resource.identifier() in git_search:
                RobustLogger().info("Found chitin/core location '%s' for resource '%s' for module '%s'", resource.filepath(), resource.identifier(), display_name)
                git_search_result_filepaths: tuple[Path, ...] = (resource.filepath(),)
                self.add_locations(resource.resname(), resource.restype(), git_search_result_filepaths).activate()

        # Prioritize Override by checking/activating last.
        for directory in self._installation.override_list():
            for resource in self._installation.override_resources(directory):
                if (
                    resource.identifier() not in self.resources  # irrelevant resources in override are skipped
                    and resource.identifier() not in git_search
                ):
                    continue
                RobustLogger().info("Found override location '%s' for module '%s'", resource.filepath(), display_name)
                self.add_locations(resource.resname(), resource.restype(), [resource.filepath()]).activate()

        # Also try get paths for textures in models
        if not self._load_textures:
            # Fast mode: skip model texture/lightmap crawling. This is the largest bottleneck in practice
            # and is unnecessary for workflows that only need module-local resources (e.g. IndoorMap/ModuleKit extraction).
            return

        lookup_texture_queries: set[str] = set()
        lookup_lightmap_queries: set[str] = set()
        for model in self.models():
            print(f"Finding textures/lightmaps for model '{model.identifier()}'...")
            try:
                model_data: bytes | None = model.data()
            except OSError:
                RobustLogger().warning(
                    "Suppressed known exception while executing %s.reload_resources() while getting model data '%s': %s",
                    repr(self),
                    model.identifier(),
                    exc_info=True,
                    extra={"detailed": False},
                )
                continue
            else:
                if model_data is None:
                    RobustLogger().warning(f"Missing model '{model.identifier()}', needed by module '{display_name}'")
                    continue
                if not model_data:
                    RobustLogger().warning(f"model '{model.identifier()}' was unexpectedly empty, but is needed by module '{display_name}'")
                    continue
            model_textures: set[str] = set()
            model_lightmaps: set[str] = set()
            with suppress(Exception):
                model_textures.update(iterate_textures(model_data))
                lookup_texture_queries.update(model_textures)
            with suppress(Exception):
                model_lightmaps.update(iterate_lightmaps(model_data))
                lookup_lightmap_queries.update(model_lightmaps)
            if model_textures:
                print(f"    Textures: {', '.join(sorted(model_textures))}")
            if model_lightmaps:
                print(f"    Lightmaps: {', '.join(sorted(model_lightmaps))}")

        texlm_queries: set[str] = lookup_texture_queries | lookup_lightmap_queries
        texture_queries: list[ResourceIdentifier] = [
            ResourceIdentifier(texture, res_type)  # create the search queries for TPC/TGA textures
            for texture in texlm_queries
            for res_type in (ResourceType.TPC, ResourceType.TGA)
        ]
        texture_search: dict[ResourceIdentifier, list[LocationResult]] = self._installation.locations(
            texture_queries,
            [
                SearchLocation.OVERRIDE,
                SearchLocation.CHITIN,
                SearchLocation.TEXTURES_TPA,  # tpa is the highest quality texture location (rather than tpb/tpc)
                SearchLocation.TEXTURES_TPB,
                SearchLocation.TEXTURES_TPC,
            ],
        )
        for identifier, locations in texture_search.items():
            if not locations:
                continue
            location_paths = [str(loc.filepath) for loc in locations]
            if len(location_paths) <= 3:
               paths_str = ', '.join(location_paths)
            else:
               paths_str = ', '.join(location_paths[:3]) + f', ... and {len(location_paths) - 3} more'
            RobustLogger().debug(f"Adding {len(locations)} texture location(s) for '{identifier.resname}.{identifier.restype.extension}' to '{display_name}': {paths_str}")
            self.add_locations(identifier.resname, identifier.restype, (location.filepath for location in locations)).activate()

        # Finally iterate through all resources we may have missed.
        for ident, module_resource in self.resources.items():
            if module_resource.isActive():
                continue
            if ident.restype == ResourceType.TPC and ResourceIdentifier(ident.resname, ResourceType.TGA) in self.resources:
                continue  # Skip TPC resources if the TGA equivalent resource is already found and activated.
            if ident.restype == ResourceType.TGA and ResourceIdentifier(ident.resname, ResourceType.TPC) in self.resources:
                continue  # Skip TGA resources if the TPC equivalent resource is already found and activated.
            module_resource.activate()

    def _handle_git_lyt_reloads(
        self,
        main_search_results: dict[ResourceIdentifier, list[LocationResult]],
        query: ResourceIdentifier,
        useable_type: type[GIT | LYT | VIS],
        errmsg: str,
    ) -> set[ResourceIdentifier]:
        if not main_search_results.get(query):
            if useable_type == VIS:
                return set()  # make vis optional I guess
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self.lookup_main_capsule().filepath() / str(query))
        original_git_or_lyt = self.add_locations(
            query.resname,
            query.restype,
            (loc.filepath for loc in main_search_results[query]),
        )
        # Activate each GIT/LYT location for this module, and fill this module with all of their resources (all of the resources their instances point to).
        original_path: Path = original_git_or_lyt.locations()[0]
        result: set[ResourceIdentifier] = set()
        for location in original_git_or_lyt.locations():
            original_git_or_lyt.activate(location)
            try:
                loaded_git_or_lyt: type[GIT | LYT | VIS] | None = original_git_or_lyt.resource()
                if loaded_git_or_lyt is None:
                    RobustLogger().warning("Failed to load resource '%s' from location '%s'", original_git_or_lyt.identifier(), location)
                    if useable_type != VIS:
                        raise RuntimeError(errmsg)  # noqa: TRY004, TRY301
                    continue  # VIS is optional, so we can skip if it fails to load
                if isinstance(loaded_git_or_lyt, VIS):
                    RobustLogger().debug("Loaded VIS resource '%s' from location '%s'", original_git_or_lyt.identifier(), location)
                    # VIS files don't have resource identifiers to iterate, so just skip
                    continue
                if isinstance(loaded_git_or_lyt, (GIT, LYT)):
                    result.update(loaded_git_or_lyt.iter_resource_identifiers())
                else:
                    RobustLogger().error(
                        "Unexpected resource type '%s' for '%s' (expected GIT, LYT, or VIS)",
                        type(loaded_git_or_lyt).__name__,
                        original_git_or_lyt.identifier(),
                    )
                    if useable_type != VIS:
                        raise RuntimeError(errmsg)  # noqa: TRY004, TRY301
            except RuntimeError:
                raise
            except Exception:  # noqa: BLE001
                RobustLogger().error("Unexpected exception when executing %s._handle_git_lyt_reloads() with resource '%s'", repr(self), original_git_or_lyt.identifier())
        original_git_or_lyt.activate(original_path)  # reactivate the main one.

        return result

    def add_locations(
        self,
        resname: str,
        restype: ResourceType,
        locations: Iterable[Path],
    ) -> ModuleResource:
        """Creates or extends a ModuleResource keyed by the resname/restype with additional locations.

        This is how Module.resources dict gets filled.

        Args:
        ----
            resname: The resource name.
            restype: The resource type.
            locations: The locations of the resource files.

        Processing Logic:
        ----------------
            - Checks if the resource already exists in the dictionary
            - If it doesn't exist, creates a new ModuleResource object
            - Adds the locations to the existing or newly created ModuleResource
            - Does not return anything, modifies the dictionary in-place.
        """
        if not isinstance(locations, Collection):
            locations = list(locations)
        if (
            not locations  # vvv skip dirt.tpc, some constant from the model-ascii data vvv
            and (resname != "dirt" or restype != ResourceType.TPC)
        ):
            RobustLogger().warning("No locations found for '%s.%s' which are intended to add to module '%s'", resname, restype.extension, self._root)
        module_resource: ModuleResource | None = self.resource(resname, restype)
        if module_resource is None:
            module_resource = ModuleResource(resname, restype, self._installation, self._root)
            self.resources[module_resource.identifier()] = module_resource
        module_resource.add_locations(locations)
        return module_resource

    def installation(self) -> Installation:
        return self._installation

    def resource(
        self,
        resname: str,
        restype: ResourceType,
    ) -> ModuleResource | None:
        """Returns the resource with the given name and type from the module.

        Args:
        ----
            resname (str): The name of the resource.
            restype (ResourceType): The type of the resource.

        Returns:
        -------
            ModuleResource | None: The resource with the given name and type, or None if it does not exist.
        """
        ident = ResourceIdentifier(resname, restype)
        return self.resources.get(ident, None)

    def layout(self) -> ModuleResource[LYT] | None:
        """Returns the LYT layout resource with a matching ID if it exists.

        Args:
        ----
            self: The Module instance

        Returns:
        -------
            ModuleResource[LYT] | None: The layout resource or None if not found

        Processing Logic:
        ----------------
            - Iterates through all resources in self.resources
            - Checks if resource name matches self._id and type is LYT
            - Returns first matching resource or None if not found.
        """
        return next(
            (resource for resource in self.resources.values() if (resource.restype() == ResourceType.LYT and resource.resname() == self.module_id())),
            None,
        )

    def vis(self) -> ModuleResource[VIS] | None:
        """Finds the VIS resource with matching ID.

        Args:
        ----
            self: The Module object.

        Returns:
        -------
            ModuleResource[VIS] | None: The VIS resource object or None.

        Finds the VIS resource object from the Module's resources:
            - Iterates through the resources dictionary values
            - Checks if the resource name matches self._id in lowercase and type is VIS
            - Returns the first matching resource or None.
        """
        return next((resource for resource in self.resources.values() if (resource.restype() == ResourceType.VIS and resource.resname() == self.module_id())), None)

    def are(self) -> ModuleResource[ARE] | None:
        """Returns the ARE resource with the given ID if it exists.

        Args:
        ----
            self: The Module object

        Returns:
        -------
            ModuleResource[ARE] | None: The ARE resource or None if not found

        Processing Logic:
        ----------------
            - Iterate through all resources in self.resources
            - Check if resource name matches self._id in lowercase and resource type is ARE
            - Return first matching resource or None if no match.
        """
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.ARE and resource.resname() == self.module_id()),
            None,
        )

    def git(self) -> ModuleResource[GIT] | None:
        """Returns the git resource with matching id if found.

        Args:
        ----
            self: The module object

        Returns:
        -------
            ModuleResource[GIT] | None: The git resource or None

        Processing Logic:
        ----------------
            - Iterate through all resources in module
            - Check if resource name matches id in lowercase and type is GIT
            - Return matching resource or None if not found.
        """
        result = next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.GIT and resource.resname() == self.module_id()),
            None,
        )
        if result is None:  # noqa: RET503
            fallback = next(
                (resource for resource in self.resources.values() if resource.restype() == ResourceType.GIT),
                None,
            )
            if fallback is not None:  # noqa: RET503
                RobustLogger().warning("This module '%s' has an incorrect GIT resname/resref! Expected '%s', found '%s'", self._root, self.module_id(), fallback.resname())  # noqa: RET503
        return result  # noqa: RET504

    def pth(self) -> ModuleResource[PTH] | None:
        """Finds the PTH resource with matching ID.

        Args:
        ----
            self: The Module object.

        Returns:
        -------
            ModuleResource[PTH] | None: The PTH resource or None if not found.

        Finds the PTH resource:
            - Iterates through all resources
            - Checks if resource name matches self._id and type is PTH
            - Returns first matching resource or None.
        """
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.PTH and resource.resname() == self.module_id()),
            None,
        )

    def ifo(self) -> ModuleResource[IFO] | None:
        return self.info()

    def info(self) -> ModuleResource[IFO] | None:
        """Returns the ModuleResource with type IFO if it exists.

        Args:
        ----
            self: The object instance

        Returns:
        -------
            ModuleResource[IFO] | None: The ModuleResource with type IFO or None

        Processing Logic:
        ----------------
            - Iterate through self.resources values
            - Check if resource name is 'module' and type is IFO
            - Return first matching resource
            - Return None if no match found.
        """
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.IFO and resource.identifier().lower_resname == "module"),
            None,
        )

    def creature(
        self,
        resname: str,
    ) -> ModuleResource[UTC] | None:
        """Returns a UTC resource by name if it exists.

        Args:
        ----
            resname: Name of the resource to search for

        Returns:
        -------
            ModuleResource[UTC]: The UTC resource or None if not found

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary values
            - Check if resname matches resource name and type is UTC
            - Return matching resource or None if not found.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTC and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def _filter_resources_by_type(self, resource_type: ResourceType) -> list[ModuleResource]:
        """Filter resources by their type.

        Args:
        ----
            resource_type: The ResourceType to filter by

        Returns:
        -------
            list[ModuleResource]: List of resources matching the specified type

        Processing Logic:
        ----------------
            - Iterate through all resources in self.resources.values()
            - Check if each resource's type matches the specified resource_type
            - Return list of matching resources.
        """
        return [resource for resource in self.resources.values() if resource.restype() == resource_type]

    def _filter_resources_by_types(self, resource_types: set[ResourceType], require_active: bool = False) -> list[ModuleResource]:
        """Filter resources by their types with optional active check.

        Args:
        ----
            resource_types: Set of ResourceTypes to filter by
            require_active: Whether to only include active resources

        Returns:
        -------
            list[ModuleResource]: List of resources matching the specified types

        Processing Logic:
        ----------------
            - Iterate through all resources in self.resources.values()
            - Check if each resource's type is in the specified resource_types
            - If require_active is True, also check if resource.isActive()
            - Return list of matching resources.
        """
        return [
            resource
            for resource in self.resources.values()
            if resource.restype() in resource_types and (not require_active or resource.isActive())
        ]

    def creatures(self) -> list[ModuleResource[UTC]]:
        """Returns a list of UTC resources.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[UTC]]: A list of UTC resources

        Processing Logic:
        ----------------
            - Iterate through all resources in self.resources
            - Check if each resource's type is UTC
            - Add matching resources to the return list.
        """
        return self._filter_resources_by_type(ResourceType.UTC)

    def placeable(
        self,
        resname: str,
    ) -> ModuleResource[UTP] | None:
        """Check if a placeable UTP resource with the given resname exists.

        Args:
        ----
            resname (str): Name of the resource to check

        Returns:
        -------
            resource: Found resource or None

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary
            - Check if resource name matches given name and type is UTP
            - Return matching resource if found, else return None.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTP and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def placeables(self) -> list[ModuleResource[UTP]]:
        """Returns a list of UTP resources for this module.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[UTP]]: List of UTP resources

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary
            - Check if resource type is UTP
            - Add matching resources to the return list.
        """
        return self._filter_resources_by_type(ResourceType.UTP)

    def door(
        self,
        resname: str,
    ) -> ModuleResource[UTD] | None:
        """Returns a UTD resource matching the provided resname from this module.

        Args:
        ----
            resname (str): The name of the resource

        Returns:
        -------
            ModuleResource[UTD] | None: The UTD resource or None if not found

        Processing Logic:
        ----------------
            - Iterate through self.resources values
            - Check if resname matches resource name and type is UTD
            - Return matching resource or None if not found.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTD and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def doors(self) -> list[ModuleResource[UTD]]:
        """Returns a list of all UTD resources for this module.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[UTD]]: List of UTD resources

        Processing Logic:
        ----------------
            - Iterate through all resources stored in self.resources
            - Check if each resource's type is UTD
            - Add matching resources to the return list.
        """
        return self._filter_resources_by_type(ResourceType.UTD)

    def item(
        self,
        resname: str,
    ) -> ModuleResource[UTI] | None:
        """Returns a UTI resource matching the provided resname from this module if it exists.

        Args:
        ----
            resname (str): Name of the resource to lookup

        Returns:
        -------
            ModuleResource[UTI] | None: The matching UTI resource or None

        Processing Logic:
        ----------------
            - Iterates through self.resources dictionary values
            - Returns the first resource where resname matches resource.resname() and resource type is UTI
            - Returns None if no matching resource found.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTI and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def items(self) -> list[ModuleResource[UTI]]:
        """Returns a list of UTI resources for this module.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[UTI]]: A list of UTI resources

        Processing Logic:
        ----------------
            - Iterate through self.resources which is a dictionary of all resources
            - Check if each resource's restype is equal to ResourceType.UTI
            - If equal, add it to the return list
            - Return the list of UTI resources.
        """
        return self._filter_resources_by_type(ResourceType.UTI)

    def encounter(
        self,
        resname: str,
    ) -> ModuleResource[UTE] | None:
        """Find UTE resource by the specified resname.

        Args:
        ----
            resname: Resource name to search for

        Returns:
        -------
            resource: Found UTE resource or None

        Processing Logic:
        ----------------
            - Iterate through self.resources values
            - Check if resname matches resource name and type is UTE
            - Return first matching resource or None.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTE and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def encounters(self) -> list[ModuleResource[UTE]]:
        """Returns a list of UTE resources for this module.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[UTE]]: A list of UTE resources

        Processing Logic:
        ----------------
            - Iterate through all resources stored in self.resources
            - Check if each resource's type is UTE
            - If type matches, add it to the return list
            - Return the list of UTE resources.
        """
        return self._filter_resources_by_type(ResourceType.UTE)

    def store(self, resname: str) -> ModuleResource[UTM] | None:
        """Looks up a material (UTM) resource by the specified resname from this module and returns the resource data.

        Args:
        ----
            resname(str): Name of the resource to look up

        Returns:
        -------
            resource: The looked up resource or None if not found

        Processing Logic:
        ----------------
            - Loops through all resources stored in self.resources
            - Checks if the resource name matches the given name and type is UTM
            - Returns the first matching resource
            - Returns None if no match found.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTM and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def stores(self) -> list[ModuleResource[UTM]]:
        """Returns a list of material (UTM) resources for this module."""
        return self._filter_resources_by_type(ResourceType.UTM)

    def trigger(
        self,
        resname: str,
    ) -> ModuleResource[UTT] | None:
        """Returns a trigger (UTT) resource by the specified resname if it exists.

        Args:
        ----
            resname: Name of the resource to retrieve

        Returns:
        -------
            resource: The requested UTT resource or None

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary values
            - Check if resname matches resource name and type is UTT
            - Return first matching resource
            - Return None if no match found.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTT and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def triggers(self) -> list[ModuleResource[UTT]]:
        """Returns a list of UTT resources for this module.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[UTT]]: A list of UTT resources

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary
            - Check if each resource's restype is UTT
            - Add matching resources to a list
            - Return the list of UTT resources.
        """
        return self._filter_resources_by_type(ResourceType.UTT)

    def waypoint(
        self,
        resname: str,
    ) -> ModuleResource[UTW] | None:
        """Returns the UTW resource with the given name if it exists.

        Args:
        ----
            resname: The name of the UTW resource

        Returns:
        -------
            resource: The UTW resource or None if not found

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary values
            - Check if resname matches resource name and type is UTW
            - Return first matching resource
            - Return None if no match found.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTW and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def waypoints(self) -> list[ModuleResource[UTW]]:
        """Returns list of UTW resources from resources dict.

        Returns:
        -------
            list[ModuleResource[UTW]]: List of UTW resources

        Processing Logic:
        ----------------
            - Iterate through self.resources dict values
            - Check if resource type is UTW
            - Add matching resources to return list
            - Return list of UTW resources.
        """
        return self._filter_resources_by_type(ResourceType.UTW)

    def model(
        self,
        resname: str,
    ) -> ModuleResource[MDL] | None:
        """Returns a ModuleResource object for the given resource name if it exists in this module.

        Args:
        ----
            resname: The name of the resource to lookup.

        Returns:
        -------
            resource: The ModuleResource object if found, None otherwise.

        Processing Logic:
        ----------------
            - Loops through all resources stored in self.resources
            - Checks if the resource name matches the given name and the resource type is MDL
            - Returns the matching resource if found, None otherwise.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.MDL and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def model_ext(
        self,
        resname: str,
    ) -> ModuleResource | None:
        """Finds a MDX module resource by name from this module.

        Args:
        ----
            resname: The name of the resource to find.

        Returns:
        -------
            ModuleResource|None: The matching resource or None if not found.

        Processes the resources dictionary:
            - Iterates through resources.values()
            - Checks if resname matches resource.resname() and resource type is MDX
            - Returns first matching resource or None.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.MDX and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def models(self) -> list[ModuleResource[MDL]]:
        """Returns a list of MDL model resources.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[MDL]]: A list of MDL model resources

        Processes the resources dictionary:
            - Loops through each value in the resources dictionary
            - Checks if the resource type is MDL
            - Adds matching resources to the return list.
        """
        return self._filter_resources_by_type(ResourceType.MDL)

    def model_exts(self) -> list[ModuleResource]:
        """Returns a list of MDX model resources.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource]: A list of MDX model resources

        Processes the resources dictionary:
            - Loops through each value in the resources dictionary
            - Checks if the resource type is MDX
            - Adds matching resources to the return list.
        """
        return self._filter_resources_by_type(ResourceType.MDX)

    def texture(
        self,
        resname: str,
    ) -> ModuleResource[TPC] | None:
        """Looks up a texture resource by resname from this module.

        Args:
        ----
            resname: Name of the texture resource to look up.

        Returns:
        -------
            resource: Found texture resource or None.

        Processing Logic:
        ----------------
            - Loops through all resources stored in self.resources
            - Checks if resname matches the resource name in any case-insensitive way
            - Checks if the resource type is a texture format like TPC or TGA
            - Returns the first matching resource or None if not found.
        """
        lower_resname: str = resname.lower()
        texture_types: set[ResourceType] = {ResourceType.TPC, ResourceType.TGA}
        return next(
            (
                resource
                for resource in self.resources.values()
                if resource.isActive() and resource.restype() in texture_types and lower_resname == resource.identifier().lower_resname
            ),
            None,
        )

    def textures(self) -> list[ModuleResource[MDL]]:
        """Generates a list of texture resources from this module.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[MDL]]: List of texture resources

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary
            - Check if resource type is TPC or TGA texture format
            - Include the resource in return list if type matches.
        """
        texture_types: set[ResourceType] = {ResourceType.TPC, ResourceType.TGA}
        return self._filter_resources_by_types(texture_types, require_active=True)

    def sound(
        self,
        resname: str,
    ) -> ModuleResource[UTS] | None:
        """Returns the UTS resource with the given name if it exists.

        Args:
        ----
            resname: The name of the UTS resource

        Returns:
        -------
            resource: The UTS resource or None if not found

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary values
            - Check if resname matches resource name and type is UTS
            - Return matching resource or None if not found.
        """
        lower_resname: str = resname.lower()
        return next(
            (resource for resource in self.resources.values() if resource.restype() == ResourceType.UTS and lower_resname == resource.identifier().lower_resname),
            None,
        )

    def sounds(self) -> list[ModuleResource[UTS]]:
        """Returns a list of UTS resources.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            list[ModuleResource[UTS]]: A list of UTS resources

        Processing Logic:
        ----------------
            - Iterate through self.resources dictionary
            - Check if each resource's type is UTS
            - Add matching resources to a list
            - Return the list of UTS resources.
        """
        return self._filter_resources_by_type(ResourceType.UTS)

    def loadscreen(self) -> FileResource | None:
        """Returns a FileResource object representing the loadscreen texture for this module.

        The loadscreen is determined by:
        1. Getting the LoadScreenID from the module's ARE file
        2. Looking up the bmpresref in loadscreens.2da using the LoadScreenID
        3. Finding the texture resource (TGA or TPC) with that ResRef

        References:
        ----------
        Based on swkotor.exe GFF/ARE structure:
        - CSWSArea::LoadAreaHeader (/K1/k1_win_gog_swkotor.exe @ 0x00508c50, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Loads area header from GFF
        - C2DA::Load2DArray (/K1/k1_win_gog_swkotor.exe @ 0x004143b0, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Loads 2DA file from resource
        - CResGFF::CreateGFFFile (/K1/k1_win_gog_swkotor.exe @ 0x00411260, /K2/k2_win_gog_aspyr_swkotor2.exe: TODO: Find this address) - Creates GFF file structure
        wiki/2DA-loadscreens.md - loadscreens.2da structure and bmpresref column



        Returns:
        -------
            FileResource | None: The loadscreen texture FileResource, or None if not found.

        Processing Logic:
        ----------------
            1. Get the ARE resource from the module
            2. Read the ARE to get LoadScreenID
            3. Load loadscreens.2da from the installation
            4. Get the bmpresref from loadscreens.2da using LoadScreenID as row index
            5. Search for the texture (TGA or TPC) using installation.locations()
            6. Return the FileResource from the first location found, or None if not found.
        """
        from pykotor.resource.formats.twoda.twoda_auto import read_2da

        # Get the ARE resource
        are_resource = self.are()
        if are_resource is None:
            RobustLogger().warning(f"Module '{self._root}' has no ARE resource, cannot determine loadscreen")
            return None

        # Read the ARE to get LoadScreenID
        are_data = are_resource.resource()
        if are_data is None:
            RobustLogger().warning(f"Failed to read ARE resource for module '{self._root}'")
            return None

        loadscreen_id = are_data.loadscreen_id
        if loadscreen_id == 0:
            RobustLogger().debug(f"Module '{self._root}' has LoadScreenID=0, no loadscreen specified")
            return None

        # Load loadscreens.2da from installation
        loadscreens_result = self._installation.resource(
            "loadscreens",
            ResourceType.TwoDA,
            [SearchLocation.OVERRIDE, SearchLocation.CHITIN],
        )
        if loadscreens_result is None:
            RobustLogger().warning("loadscreens.2da not found in installation")
            return None

        loadscreens_2da = read_2da(loadscreens_result.data)

        # Get the bmpresref from loadscreens.2da using LoadScreenID as row index
        try:
            loadscreen_row = loadscreens_2da.get_row(loadscreen_id)
            bmpresref = loadscreen_row.get_string("bmpresref")
            if not bmpresref or bmpresref == "****":
                RobustLogger().debug(f"Module '{self._root}' loadscreen row {loadscreen_id} has no bmpresref")
                return None
        except (IndexError, KeyError) as e:
            RobustLogger().warning(f"Failed to get bmpresref from loadscreens.2da row {loadscreen_id}: {e}")
            return None

        # Search for the texture (TGA or TPC) using installation.locations()
        texture_queries = [
            ResourceIdentifier(bmpresref, ResourceType.TPC),
            ResourceIdentifier(bmpresref, ResourceType.TGA),
        ]
        texture_locations = self._installation.locations(
            texture_queries,
            [SearchLocation.OVERRIDE, SearchLocation.CUSTOM_MODULES, SearchLocation.CHITIN, SearchLocation.TEXTURES_TPA],
        )

        # Return the FileResource from the first location found
        for query in texture_queries:
            if query in texture_locations and texture_locations[query]:
                location: LocationResult = texture_locations[query][0]
                try:
                    return location.as_file_resource()
                except RuntimeError:
                    # If FileResource wasn't set, create one from the location
                    return FileResource(
                        resname=bmpresref,
                        restype=query.restype,
                        size=location.size,
                        offset=location.offset,
                        filepath=location.filepath,
                    )

        RobustLogger().debug(f"Loadscreen texture '{bmpresref}' not found for module '{self._root}'")
        return None


class ModuleResource(Generic[T]):
    """Represents a single resource within a module with multiple possible locations.

    ModuleResource manages a resource that may exist in multiple locations (override,
    module archives, chitin). It tracks all locations and allows activation of a
    specific location, with lazy loading of the actual resource object.

    References:
    ----------
        Original BioWare Odyssey Engine (resource search order: Override > Module > Chitin from swkotor.exe, swkotor2.exe)



    Attributes:
    ----------
        _resname: Resource name (ResRef) without extension.
            The name of the resource (e.g., "module", "danm13").

        _restype: Resource type identifier.
            The type of resource (e.g., ResourceType.IFO, ResourceType.ARE).

        _installation: Installation instance for resource lookups.
            Used to resolve resources from chitin and other locations.

        _active: Currently active file path for this resource.
            The file path currently being used to load this resource.
            None if no location has been activated yet.

        _resource_obj: Cached loaded resource object.
            The parsed resource object (e.g., IFO, ARE, UTC).
            None until resource() is called for the first time.

        _locations: List of all file paths where this resource exists.
            All known locations for this resource, ordered by priority.
            Search order: Override > Custom Modules > Chitin

        _identifier: ResourceIdentifier for this resource.
            Reference: PyKotor-specific abstraction
            Combines resname and restype for unique identification.
    """

    def __init__(
        self,
        resname: str,
        restype: ResourceType,
        installation: Installation,
        module_root: str | None = None,
    ):
        self._resname: str = resname
        self._installation: Installation = installation
        self._restype: ResourceType = restype
        self._active: Path | None = None
        self._resource_obj: Any = None
        self._locations: list[Path] = []
        self._identifier = ResourceIdentifier(resname, restype)
        self._module_root: str | None = module_root

    def __repr__(self):
        return f"{self.__class__.__name__}(resname={self._resname} restype={self._restype!r} installation={self._installation!r})"

    def __eq__(self, other):
        if self is other:
            return True
        if isinstance(other, ResourceIdentifier):
            return self._identifier == other
        if isinstance(other, ModuleResource):
            return self._identifier == other._identifier
        return NotImplemented  # type: ignore[no-any-return]

    def __hash__(self):
        return hash(self._identifier)

    def resname(self) -> str:
        """Returns the resource name.

        Returns:
        -------
            The resource name.
        """
        return self._resname

    def restype(self) -> ResourceType:
        """Returns the type of resource stored.

        Returns:
        -------
            The resource type.
        """
        return self._restype

    def filename(self) -> str:
        return str(self._identifier)

    def identifier(self) -> ResourceIdentifier:
        return self._identifier

    def localized_name(self) -> str | None:
        # sourcery skip: assign-if-exp, reintroduce-else
        """Returns a localized name for the resource.

        Args:
        ----
            self: The object instance

        Returns:
        -------
            str | None: Localized name or None if not found

        Processing Logic:
        ----------------
            - Get the resource from self.resource()
            - Check if resource is None and return None
            - Check type of resource and return localized name by calling installation string method
            - Return None if type is not matched.
        """
        res: T | None = self.resource()
        if res is None:
            return None
        if isinstance(res, UTC):
            return f"{self._installation.string(res.first_name)} {self._installation.string(res.last_name)}"
        if isinstance(res, (UTD, UTE, UTM, UTP, UTS, UTT, UTW)):
            return self._installation.string(res.name)
        print(f"Could not find a localized name for a ModuleResource typed {type(res).__name__}")
        return None

    def data(self) -> bytes | None:
        """Opens the file at the active location and returns the data.

        Raises:
        ------
            ValueError: If no file is active.

        Returns:
        -------
            The bytes data of the active file.
        """
        file_name = f"{self._resname}.{self._restype.extension}"
        active_path = self.active()
        if active_path is None:
            return None

        if is_capsule_file(active_path):
            data: bytes | None = Capsule(active_path).resource(self._resname, self._restype)
            if data is None:
                RobustLogger().error(f"Resource '{file_name}' not found in '{active_path}'")
            return data

        if is_bif_file(active_path):
            resource: ResourceResult | None = self._installation.resource(
                self._resname,
                self._restype,
                [SearchLocation.CHITIN],
            )
            if resource is None:
                msg = f"Resource '{file_name}' not found in BIF '{self._active}' somehow?"
                RobustLogger().error(msg)
                return None
            return resource.data

        return active_path.read_bytes()

    def resource(self) -> T | None:
        """Returns the cached resource object. If no object has been cached, then it will load the object.

        Returns:
        -------
            The resource object.

        Returns:
            The abstracted resource, or None if not found.
        """
        if self._resource_obj is None:
            conversions: dict[ResourceType, Callable[[SOURCE_TYPES], Any]] = {
                ResourceType.ARE: read_are,
                ResourceType.DLG: read_dlg,
                ResourceType.GIT: read_git,
                ResourceType.IFO: read_ifo,
                ResourceType.LYT: read_lyt,
                ResourceType.NCS: lambda data: data,
                ResourceType.PTH: read_pth,
                ResourceType.TPC: read_tpc,
                ResourceType.TGA: read_tpc,
                ResourceType.UTD: read_utd,
                ResourceType.UTE: read_ute,
                ResourceType.UTI: read_uti,
                ResourceType.UTM: read_utm,
                ResourceType.UTP: read_utp,
                ResourceType.UTS: read_uts,
                ResourceType.UTT: read_utt,
                ResourceType.UTW: read_utw,
                ResourceType.UTC: read_utc,
                ResourceType.VIS: read_vis,
                ResourceType.WOK: read_bwm,
            }
            active_path = self.active()
            if active_path is None:
                return None

            if is_capsule_file(active_path):
                data: bytes | None = Capsule(active_path).resource(self._resname, self._restype)
                if data is None:
                    msg = f"Resource '{self._identifier}' not found in '{active_path}'"
                    RobustLogger().error(msg)
                    return None
                try:
                    self._resource_obj = conversions.get(self._restype, lambda _: None)(data)
                except Exception as e:  # noqa: BLE001
                    # Some shipped installations contain malformed VIS. Treat it as missing so
                    # module workflows that do not require VIS (e.g. walkmesh extraction/build)
                    # can still function.
                    if self._restype == ResourceType.VIS:
                        RobustLogger().warning(
                            "Failed to parse VIS '%s.%s' from '%s': %s",
                            self._resname,
                            self._restype.extension,
                            active_path,
                            e,
                        )
                        self._resource_obj = None
                    else:
                        raise

            elif is_bif_file(active_path):
                resource: ResourceResult | None = self._installation.resource(
                    self._resname,
                    self._restype,
                    [SearchLocation.CHITIN],
                )
                if resource is None:
                    msg = f"Resource '{self._identifier}' not found in '{active_path}'"
                    RobustLogger().error(msg)
                    return None
                try:
                    self._resource_obj = conversions.get(self._restype, lambda _: None)(resource.data)
                except Exception as e:  # noqa: BLE001
                    if self._restype == ResourceType.VIS:
                        RobustLogger().warning(
                            "Failed to parse VIS '%s.%s' from '%s': %s",
                            self._resname,
                            self._restype.extension,
                            active_path,
                            e,
                        )
                        self._resource_obj = None
                    else:
                        raise

            else:
                data = active_path.read_bytes()
                try:
                    self._resource_obj = conversions.get(self._restype, lambda _: None)(data)
                except Exception as e:  # noqa: BLE001
                    if self._restype == ResourceType.VIS:
                        RobustLogger().warning(
                            "Failed to parse VIS '%s.%s' from '%s': %s",
                            self._resname,
                            self._restype.extension,
                            active_path,
                            e,
                        )
                        self._resource_obj = None
                    else:
                        raise

        return self._resource_obj

    def to_bytes(self) -> bytes | None:
        if self._resource_obj is None:
            return None
        conversions: dict[ResourceType, Callable[[Any, TARGET_TYPES], Any]] = {
            ResourceType.ARE: write_are,
            #            ResourceType.CNV: write_cnv,
            ResourceType.DLG: write_dlg,
            ResourceType.DWK: write_bwm,
            #            ResourceType.FAC: write_fac,
            ResourceType.GIT: write_git,
            ResourceType.IFO: write_ifo,
            ResourceType.LYT: write_lyt,
            #            ResourceType.MDL: write_mdl,
            #            ResourceType.MDX: write_mdl,
            ResourceType.NCS: write_ncs,
            ResourceType.PTH: write_pth,
            #            ResourceType.PWK: write_bwm,
            ResourceType.TPC: write_tpc,
            ResourceType.TGA: write_tpc,
            ResourceType.UTC: write_utc,
            ResourceType.UTD: write_utd,
            ResourceType.UTE: write_ute,
            ResourceType.UTI: write_uti,
            ResourceType.UTM: write_utm,
            ResourceType.UTP: write_utp,
            ResourceType.UTS: write_uts,
            ResourceType.UTT: write_utt,
            ResourceType.UTW: write_utw,
            ResourceType.VIS: write_vis,
            ResourceType.WOK: write_bwm,
        }
        result: bytearray = bytearray()
        conversions.get(self._restype, lambda _a, _b: b"")(self._resource_obj, result)
        return bytes(result)

    def add_locations(self, filepaths: Iterable[Path]):
        """Adds a list of filepaths to the list of locations stored for the resource.

        If a filepath already exists, it is ignored.

        Args:
        ----
            filepaths: A list of filepaths pointing to a location for the resource.
        """
        self._locations.extend(filepath for filepath in filepaths if filepath not in self._locations)

    def locations(self) -> list[Path]:
        return self._locations

    def activate(
        self,
        filepath: os.PathLike | str | None = None,
    ) -> Path | None:
        """Sets the active file to the specified path. Calling this method will reset the loaded resource.

        If the filepath is not in the stored locations, calling this method will add it.

        Args:
        ----
            filepath: The new active file.
        """
        self._resource_obj = None
        if filepath is None:
            self._active = next(iter(self._locations), None)
        else:
            r_filepath = Path(filepath)
            if r_filepath not in self._locations:
                self._locations.append(r_filepath)
            self._active = r_filepath
        if self._active is None:
            module_info = f" in module '{self._module_root}'" if self._module_root else ""
            installation_path = str(self._installation.path())
            locations_info = f"Searched locations: {[str(loc) for loc in self._locations]}." if self._locations else "No locations were added to this resource."
            RobustLogger().warning(
                f"Cannot activate module resource '{self.identifier()}'{module_info}: No locations found. Installation: {installation_path}. {locations_info}"
            )
        # else:
        #    other_locations_available = len(self._locations) - 1
        #    other_locations_available_display = f" ({other_locations_available} other locations available)" if other_locations_available else ""
        #    print(f"Activating module resource '{self.identifier()}' at filepath '{self._active}'{other_locations_available_display}")
        return self._active

    def unload(self):
        """Clears the cached resource object from memory."""
        self._resource_obj = None

    def reload(self):
        """Reloads the resource object from the active location."""
        self._resource_obj = None
        self.resource()

    def active(self) -> Path | None:
        """Returns the filepath of the currently active file for the resource.

        Returns:
        -------
            Filepath to the active resource.
        """
        if self._active is None:
            next_path = next(iter(self._locations), None)
            if next_path is None:
                RobustLogger().warning("No resource found for '%s'", self._identifier)
                return None
            self.activate()
            # raise RuntimeError(f"{self!r}.activate(filepath) must be called before use.")
        return self._active

    def isActive(self) -> bool:
        return bool(self._active)

    def save(self):
        """Saves the resource to the active file.

        Args:
        ----
            self: The resource object

        Returns:
        -------
            None: This function does not return anything

        Processing Logic:
        ----------------
            - Checks if an active file is selected
            - Checks file type and writes resource data accordingly
            - Writes resource data to ERF, RIM or binary file using appropriate conversion and writer.
        """
        conversions: dict[ResourceType, Callable[[Any], bytes]] = {
            ResourceType.ARE: bytes_are,
            ResourceType.DLG: bytes_dlg,
            ResourceType.GIT: bytes_git,
            ResourceType.IFO: bytes_ifo,
            ResourceType.LYT: bytes_lyt,
            ResourceType.NCS: lambda data: data,
            ResourceType.PTH: bytes_pth,
            ResourceType.TPC: bytes_tpc,
            ResourceType.TGA: bytes_tpc,
            ResourceType.UTD: bytes_utd,
            ResourceType.UTE: bytes_ute,
            ResourceType.UTI: bytes_uti,
            ResourceType.UTM: bytes_utm,
            ResourceType.UTP: bytes_utp,
            ResourceType.UTS: bytes_uts,
            ResourceType.UTT: bytes_utt,
            ResourceType.UTW: bytes_utw,
            ResourceType.UTC: bytes_utc,
            ResourceType.VIS: bytes_vis,
            ResourceType.WOK: bytes_bwm,
        }

        active_path = self.active()
        if not active_path:
            active_path = self._create_anew_in_override()
        if is_bif_file(active_path):
            msg = "Cannot save file to BIF."
            raise ValueError(msg)

        if is_any_erf_type_file(active_path):
            erf: ERF = read_erf(active_path)
            erf.set_data(
                self._resname,
                self._restype,
                conversions[self._restype](self.resource()),
            )
            write_erf(erf, active_path)

        elif is_rim_file(active_path):
            rim: RIM = read_rim(active_path)
            rim.set_data(
                self._resname,
                self._restype,
                conversions[self._restype](self.resource()),
            )
            write_rim(rim, active_path)

        else:
            data = conversions.get(self._restype, lambda _: b"")(self.resource())
            if not data:
                raise ValueError(f"No conversion available for resource type {self._restype}")
            active_path.write_bytes(data)

    def _create_anew_in_override(self) -> Path:
        res_data: bytes | None = self.to_bytes()
        if res_data is None:
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                self._installation.override_path().joinpath(self.filename()),
            )

        RobustLogger().warning(f"Saving ModuleResource '{self.identifier()}' to the Override folder as it does not have any other paths available...")
        result = self._installation.override_path().joinpath(self.filename())
        result.write_bytes(res_data)
        self.activate(result)
        return result



"""This module handles classes relating to editing VIS files.

VIS (Visibility) files define which rooms are visible from other rooms in a module.
This is used for occlusion culling optimization - the game engine only renders rooms
that are marked as visible from the player's current room. VIS files are ASCII text
files with a simple format: parent room names followed by indented child room names.

References:
----------
        Based on unified K1 (swkotor.exe) and TSL (swkotor2.exe) VIS structure.
        Addresses: (K1: swkotor.exe, TSL: swkotor2.exe — verify/fill TSL via REVA when available).

        - LoadVisibility / Scene::LoadVisibility — loads VIS file for area visibility culling; parses ASCII format; builds room visibility map.
          K1: 0x004568d0, TSL: TODO
        - "%s/%s.VIS" format string (VIS file path format): K1: 0x007415e8, TSL: TODO
        - ".vis" extension: K1: 0x00741604, TSL: TODO
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/VISObject.ts:33-276
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/interface/module/IVISRoom.ts
        ASCII Format:
        ------------
        Parent Room Line:
        room_name number_of_child_rooms

        Example: "room001 3"

    Child Room Lines (indented with 2 spaces):
          child_room_name
          child_room_name
        Example:
          room002
          room003
          room004

    Format Rules:
        - Parent rooms start at column 0
        - Child rooms are indented with exactly 2 spaces
        - Room names are case-insensitive (stored lowercase)
        - Empty lines are ignored
        - Whitespace is trimmed from room names
"""

from __future__ import annotations

from copy import copy, deepcopy
from typing import TYPE_CHECKING, Any

from pykotor.resource.formats._base import ComparableMixin
from pykotor.resource.type import ResourceType

if TYPE_CHECKING:
    from collections.abc import Generator


class VIS(ComparableMixin):
    """Represents a VIS (Visibility) file defining room visibility relationships.

    VIS files optimize rendering by specifying which rooms are visible from each
    parent room. When the player is in a room, only rooms marked as visible in
    the VIS file are rendered. This prevents rendering rooms that are occluded
    by walls or geometry, improving performance.

    References:
    ----------
        Based on swkotor.exe VIS structure:
        - Scene::LoadVisibility @ 0x004568d0 - Loads VIS file for area visibility culling
          * Parses ASCII VIS file format
          * Builds room visibility map for occlusion culling
          * Stores parent-child room relationships
        - "%s/%s.VIS" format string @ 0x007415e8 - VIS file path format
        - ".vis" extension @ 0x00741604 - VIS file extension identifier

        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/VISObject.ts:34 (rooms Map)



    Attributes:
    ----------
        _rooms: Set of all room names defined in this VIS file
            Reference: https://github.com/th3w1zard1/KotOR.js/tree/master/VISObject.ts:34,118 (rooms Map, room.name)
            Room names are stored lowercase for case-insensitive comparison
            Each room name corresponds to a room model/area in the module
            Used to validate room existence before setting visibility

        _visibility: Dictionary mapping observer rooms to sets of visible rooms
            Reference: https://github.com/th3w1zard1/KotOR.js/tree/master/VISObject.ts:99 (currentRoom.rooms array)
            Key: Observer room name (room player is currently in)
            Value: Set of room names visible from the observer room
            If room A is in _visibility[room B], then room A is visible from room B
            Used by game engine for occlusion culling optimization
    """

    BINARY_TYPE = ResourceType.VIS
    COMPARABLE_SET_FIELDS = ("_rooms",)
    COMPARABLE_FIELDS = ("_visibility",)

    def __init__(self):
        # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/VISObject.ts:34,118
        # Set of all room names (stored lowercase for case-insensitive comparison)
        self._rooms: set[str] = set()

        # https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/VISObject.ts:99
        # Dictionary: observer room -> set of visible rooms
        # Used for occlusion culling (only render visible rooms)
        self._visibility: dict[str, set[str]] = {}

    def __eq__(self, other):
        if not isinstance(other, VIS):
            return NotImplemented  # type: ignore[no-any-return]
        return self._rooms == other._rooms and self._visibility == other._visibility

    def __hash__(self):
        return hash(
            (
                tuple(sorted(self._rooms)),
                tuple(sorted((k, tuple(sorted(v))) for k, v in self._visibility.items())),
            )
        )

    def __iter__(self) -> Generator[tuple[str, set[str]], Any, None]:
        for observer, observed in self._visibility.items():
            yield observer, deepcopy(observed)

    def all_rooms(self) -> set[str]:
        """Returns a copy of the set of rooms.

        Args:
        ----
            self: The class instance

        Returns:
        -------
            set[str]: A copy of the set of rooms

        Processing Logic:
        ----------------
            - Creates a copy of the internal _rooms set to avoid direct manipulation of the original set
            - The copy() method is used to make a shallow copy of the set
            - This allows returning the rooms set without allowing external modification of the internal state
            - Returns the copy of the rooms set.
        """
        return copy(self._rooms)

    def add_room(
        self,
        model: str,
    ):
        """Adds a room. If an room already exists, it is ignored; no error is thrown.

        Args:
        ----
            model: The name or model of the room.
        """
        model = model.lower()

        if model not in self._rooms:
            self._visibility[model] = set()

        self._rooms.add(model)

    def remove_room(
        self,
        model: str,
    ):
        """Removes a room. If a room does not exist, it is ignored; no error is thrown.

        Args:
        ----
            model: The name or model of the room.
        """
        lower_model: str = model.lower()

        for room in self._rooms:
            if lower_model in self._visibility[room]:
                self._visibility[room].remove(lower_model)

        if lower_model in self._rooms:
            self._rooms.remove(lower_model)

    def rename_room(
        self,
        old: str,
        new: str,
    ):
        """Renames a room.

        Args:
        ----
            old (str): Old room name
            new (str): New room name

        Processing Logic:
        ----------------
            - Lowercase old and new room names
            - Check if old and new names are the same
            - Remove old room name from rooms list
            - Add new room name to rooms list
            - Copy visibility of old room to new room
            - Delete visibility of old room
            - Replace old room name with new in other rooms' visibility lists.
        """
        old = old.lower()
        new = new.lower()

        if old == new:
            return

        self._rooms.remove(old)
        self._rooms.add(new)

        self._visibility[new] = copy(self._visibility[old])
        del self._visibility[old]

        for other in self._visibility:
            if other != new and old in self._visibility[other]:
                self._visibility[other].remove(old)
                self._visibility[other].add(new)

    def room_exists(
        self,
        model: str,
    ) -> bool:
        """Returns true if the specified room exists.

        Returns:
        -------
            True if the room exists.
        """
        model.lower()
        return model in self._rooms

    def set_visible(
        self,
        when_inside: str,
        show: str,
        visible: bool,
    ):
        """Sets the visibility of a specified room based off when viewing from another specified room.

        Args:
        ----
            when_inside: The room of the observer.
            show: The observed room.
            visible: If the observed room is visible.
        """
        when_inside = when_inside.lower()
        show = show.lower()

        if when_inside not in self._rooms or show not in self._rooms:
            msg = "One of the specified rooms does not exist."
            raise ValueError(msg)

        if visible:
            self._visibility[when_inside].add(show)
        elif show in self._visibility[when_inside]:
            self._visibility[when_inside].remove(show)

    def get_visible(
        self,
        when_inside: str,
        show: str,
    ) -> bool:
        """Returns true if the observed room is visible from the observing room.

        Args:
        ----
            when_inside: The room of the observer.
            show: The observed room.

        Returns:
        -------
            True if the room is visible.
        """
        when_inside = when_inside.lower()
        show = show.lower()

        if when_inside not in self._rooms or show not in self._rooms:
            msg = "One of the specified rooms does not exist."
            raise ValueError(msg)

        return show in self._visibility[when_inside]

    def set_all_visible(self):
        """Sets all rooms visible from each other.

        Processing Logic:
        ----------------
            - Loop through each room in self._rooms
            - For that room, loop through all other rooms
            - Set the visibility between the current room and other room to True.
        """
        for when_inside in self._rooms:
            for show in (room for room in self._rooms if room != when_inside):
                self.set_visible(when_inside, show, visible=True)



"""Windows registry paths and game installation detection for KotOR (Steam, GOG, disc)."""

from __future__ import annotations

import os

from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Any

from loggerplus import RobustLogger
from pykotor.common.misc import Game
from utility.misc import ProcessorArchitecture

if TYPE_CHECKING:
    import types

    from winreg import HKEYType

    from typing_extensions import Literal, Self  # pyright: ignore[reportMissingModuleSource]


KOTOR_REG_PATHS: dict[Game, dict[ProcessorArchitecture, list[tuple[str, str]]]] = {
    Game.K1: {
        ProcessorArchitecture.BIT_32: [
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\Steam App 32370", "InstallLocation"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\GOG.com\Games\1207666283", "PATH"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\BioWare\SW\KOTOR", "InternalPath"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\BioWare\SW\KOTOR", "Path"),
            #            (r"HKEY_USERS\S-1-5-21-3288518552-3737095363-3281442775-1001\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\AmazonGames/Star Wars - Knights of the Old", "InstallLocation"),
        ],
        ProcessorArchitecture.BIT_64: [
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\Steam App 32370", "InstallLocation"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\GOG.com\Games\1207666283", "PATH"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\BioWare\SW\KOTOR", "InternalPath"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\BioWare\SW\KOTOR", "Path"),
        ],
    },
    Game.K2: {
        ProcessorArchitecture.BIT_32: [
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\Steam App 208580", "InstallLocation"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\GOG.com\Games\1421404581", "PATH"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\LucasArts\KotOR2", "InternalPath"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\LucasArts\KotOR2", "Path"),
        ],
        ProcessorArchitecture.BIT_64: [
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\Steam App 208580", "InstallLocation"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\GOG.com\Games\1421404581", "PATH"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\LucasArts\KotOR2", "InternalPath"),
            (r"HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\LucasArts\KotOR2", "Path"),
        ],
    },
}


# amazon's k1 reg key can be found using the below code. Doesn't store it in HKLM for some reason.
def find_software_key(software_name: str) -> str | None:
    import winreg

    with winreg.ConnectRegistry(None, winreg.HKEY_USERS) as hkey_users:
        i = 0
        while True:
            try:
                # Enumerate through the SIDs
                sid: str = winreg.EnumKey(hkey_users, i)
                software_path: str = f"{sid}\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\{software_name}"
                with suppress(FileNotFoundError), winreg.OpenKey(hkey_users, software_path) as software_key:
                    # If this point is reached, the software is installed under this SID
                    return winreg.QueryValue(software_key, "InstallLocation")
                i += 1
            except OSError:  # noqa: PERF203
                break  # No more left to iterate through.

    return None


def resolve_reg_key_to_path(
    registry: str | int | HKEYType,
    subkey: str,
    value_name: str | None = None,
) -> str | None:
    r"""Resolve a registry key to a file system path."""
    import winreg

    try:
        if isinstance(registry, str):
            root_name, key_path = registry.split("\\", 1)
            # Dynamic module attribute access based on registry root string - legitimate use of getattr
            root_key = getattr(winreg, root_name, None)
            if root_key is None:
                return None
            value_to_lookup = subkey
        else:  # e.g. int
            # treat as HKEYType, use as-is
            root_key = registry
            if value_name is None:
                msg = "value_name must be provided when a registry handle is supplied."
                raise ValueError(msg)
            key_path = subkey
            value_to_lookup = value_name

        with winreg.OpenKey(root_key, key_path) as key:
            resolved_path, _ = winreg.QueryValueEx(key, value_to_lookup)
            return str(resolved_path or "")
    except (AttributeError, FileNotFoundError, PermissionError):
        return None


def check_reg_keys_existence_and_validity() -> tuple[list[tuple[str, str]], list[tuple[str, str]]]:
    """Check registry keys for their existence and validity against default paths."""
    import winreg

    from pathlib import WindowsPath

    from pykotor.tools.path import find_kotor_paths_from_default

    non_existent_keys = []
    invalid_path_keys = []

    default_paths = find_kotor_paths_from_default()

    # Determine the system's architecture
    arch = ProcessorArchitecture.from_os()
    for game, arch_paths in KOTOR_REG_PATHS.items():
        game_defaults = default_paths[game]

        for path, name in arch_paths[arch]:
            reg_path = resolve_reg_key_to_path(winreg.HKEY_LOCAL_MACHINE, path, name)
            if reg_path is None:
                non_existent_keys.append((path, name))
            else:
                # Convert registry path to a proper WindowsPath and check existence and if it's a default path
                reg_path_obj = WindowsPath(reg_path)
                if not reg_path_obj.exists() or all(reg_path_obj != WindowsPath(default_path) for default_path in game_defaults):
                    invalid_path_keys.append((path, name))

    return non_existent_keys, invalid_path_keys


def winreg_key(game: Game) -> list[tuple[str, str]]:
    """Returns a list of registry keys that are utilized by KOTOR.

    Attributes:
    ----------
        game: Game IntEnum - The game to lookup
        access: Access permissions for the key (see winreg module).

    Raises:
    ------
        ValueError: Not on a Windows OS.
        WinError: Most likely do not have sufficient permissions.

    Returns:
    -------
        Key object or None if no key exists.
    """
    if os.name != "nt":
        msg = "Cannot get or set registry keys on a non-Windows OS."
        raise ValueError(msg)

    return KOTOR_REG_PATHS[game][ProcessorArchitecture.from_os()]


def get_winreg_path(game: Game) -> tuple[Any, int] | None | Literal[""]:
    """(untested) Returns the specified path value in the windows registry for the given game.

    Attributes:
    ----------
        game: The game to lookup in the registry

    Raises:
    ------
        ValueError: Not on a Windows OS.
        WinError: Most likely do not have sufficient permissions.
    """
    possible_kotor_reg_paths: list[tuple[str, str]] = winreg_key(game)

    try:
        import winreg

        for key_path, subkey in possible_kotor_reg_paths:
            key: HKEYType = winreg.OpenKeyEx(winreg.HKEY_LOCAL_MACHINE, key_path, 0, winreg.KEY_READ)
            return winreg.QueryValueEx(key, subkey)
    except (FileNotFoundError, PermissionError):
        return ""


def set_winreg_path(
    game: Game,
    path: str,
) -> None:
    """(untested) Sets the kotor install folder path value in the windows registry for the given game.

    Attributes:
    ----------
        game: The game to set in the registry
        path: New path value for the game.

    Raises:
    ------
        ValueError: Not on a Windows OS.
        WinError: Most likely do not have sufficient permissions.
    """
    possible_kotor_reg_paths: list[tuple[str, str]] = winreg_key(game)

    import winreg

    for key_path, subkey in possible_kotor_reg_paths:
        key: HKEYType = winreg.CreateKeyEx(
            winreg.HKEY_LOCAL_MACHINE,
            key_path,
            0,
            winreg.KEY_SET_VALUE,
        )
        winreg.SetValueEx(key, subkey, 1, winreg.REG_SZ, path)


def create_registry_path(
    hive: HKEYType | int,
    path: str,
) -> None:  # sourcery skip: raise-from-previous-error
    """Recursively creates the registry path if it doesn't exist."""
    log = RobustLogger()
    try:
        import winreg

        current_path: str = ""
        for part in path.split("\\"):
            current_path = f"{current_path}\\{part}" if current_path else part
            try:
                winreg.CreateKey(hive, current_path)
            except PermissionError as e:
                raise PermissionError("Permission denied. Administrator privileges required.") from e  # noqa: B904, TRY003, EM101
            except Exception as e:  # pylint: disable=W0718  # noqa: BLE001
                # sourcery skip: raise-specific-error
                raise Exception(f"Failed to create registry key: {current_path}") from e  # noqa: TRY002, TRY003, EM102, B904
    except Exception:  # pylint: disable=W0718  # noqa: BLE001
        log.exception("An unexpected error occurred while creating a registry path.")


def get_retail_key(game: Game) -> str:
    if ProcessorArchitecture.from_os() == ProcessorArchitecture.BIT_64:
        return r"HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\LucasArts\KotOR2" if game.is_k2() else r"HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\BioWare\SW\KOTOR"
    return r"HKEY_LOCAL_MACHINE\SOFTWARE\LucasArts\KotOR2" if game.is_k2() else r"HKEY_LOCAL_MACHINE\SOFTWARE\BioWare\SW\KOTOR"


class SpoofKotorRegistry:
    """A context manager used to safely spoof the KOTOR 1/2 disk retail registry path temporarily."""

    def __init__(
        self,
        installation_path: os.PathLike | str,
        game: Game | None = None,
    ):
        from pykotor.extract.installation import Installation

        # Key name at the path containing the value.
        self.key: str = "Path"
        self.spoofed_path: Path = Path(installation_path).resolve()

        if game is not None:
            determined_game: Game | None = game
        else:
            determined_game = Installation.determine_game(installation_path)

        if determined_game is None:
            raise ValueError(f"Could not auto-determine the game k1 or k2 from '{installation_path}'. Try sending 'game' enum to prevent auto-detections like this.")

        # Path to the key.
        self.registry_path: str = get_retail_key(determined_game)

        # The original contents of the key. None if not existing.
        self.original_value: str | None = resolve_reg_key_to_path(self.registry_path, self.key)

    def __enter__(self) -> Self:
        if self.spoofed_path != self.original_value:
            set_registry_key_value(self.registry_path, self.key, str(self.spoofed_path))
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ):
        # Revert the registry key to its original value if it was altered
        if self.original_value is not None and self.spoofed_path != self.original_value:
            set_registry_key_value(self.registry_path, self.key, self.original_value)
        # TODO(th3w1zard1): Determine what to do if the regpath never existed, as deleting it isn't easy. Set it to ""?


def set_registry_key_value(
    full_key_path: str,
    value_name: str,
    value_data: str,
) -> None:
    """Sets a registry key value, creating the key (and its parents, if necessary).

    Args:
    ----
        - full_key_path: The full registry key path, including the hive.
        - value_name: The name of the value to set.
        - value_data: The data to set for the value.

    Raises:
    ------
        - PermissionError: PyKotor doesn't have permission to change the registry (usually fixed by running as admin).
    """
    log = RobustLogger()
    try:
        import winreg

        # Parse the hive from the full key path
        hive_name: str
        sub_key: str
        hive_name, sub_key = full_key_path.split("\\", 1)
        hive: int | None = {
            "HKEY_CLASSES_ROOT": winreg.HKEY_CLASSES_ROOT,
            "HKEY_CURRENT_USER": winreg.HKEY_CURRENT_USER,
            "HKEY_LOCAL_MACHINE": winreg.HKEY_LOCAL_MACHINE,
            "HKEY_USERS": winreg.HKEY_USERS,
            "HKEY_CURRENT_CONFIG": winreg.HKEY_CURRENT_CONFIG,
        }.get(hive_name)
        if hive is None:
            log.error("Invalid registry hive '%s'.", hive_name)
            return

        # Create the registry path
        try:
            create_registry_path(hive, sub_key)
        except PermissionError:
            raise
        except Exception:  # pylint: disable=W0718  # noqa: BLE001
            log.exception("set_registry_key_value raised an error other than the expected PermissionError")
            return
        # Open or create the key at the specified path
        with winreg.CreateKeyEx(hive, sub_key, 0, winreg.KEY_WRITE | winreg.KEY_WOW64_32KEY) as key:
            # Set the value
            winreg.SetValueEx(key, value_name, 0, winreg.REG_SZ, value_data)
            log.debug("Successfully set %s to %s at %s\\%s", value_name, value_data, hive, sub_key)
    except PermissionError:
        raise
    except Exception:  # pylint: disable=W0718  # noqa: BLE001
        log.exception("An unexpected error occured while setting the registry.")


def remove_winreg_path(game: Game):
    """(untested)."""
    possible_kotor_reg_paths: list[tuple[str, str]] = winreg_key(game)

    try:
        import winreg

        for key_path, subkey in possible_kotor_reg_paths:
            key: HKEYType = winreg.OpenKeyEx(
                winreg.HKEY_LOCAL_MACHINE,
                key_path,
                0,
                winreg.KEY_SET_VALUE,
            )
            winreg.DeleteValue(key, subkey)
    except FileNotFoundError:
        ...
